@@ -31,6 +31,17 @@
 #include "dtmf.h"
 
 
+struct fragment_key {
+	str call_id;
+	str from_tag;
+};
+struct sdp_fragment {
+	struct ng_buffer *ngbuf;
+	struct timeval received;
+	GQueue streams;
+	struct sdp_ng_flags flags;
+};
+
 static pcre *info_re;
 static pcre_extra *info_ree;
 static pcre *streams_re;
@@ -39,10 +50,14 @@ static pcre_extra *streams_ree;
 int trust_address_def;
 int dtls_passive_def;
 
+static mutex_t sdp_fragments_lock;
+static GHashTable *sdp_fragments;
+
 
 INLINE int call_ng_flags_prefix(struct sdp_ng_flags *out, str *s_ori, const char *prefix,
 		void (*cb)(struct sdp_ng_flags *, str *, void *), void *ptr);
 static void call_ng_flags_str_ht(struct sdp_ng_flags *out, str *s, void *htp);
+static void ng_stats_ssrc(bencode_item_t *dict, struct ssrc_hash *ht);
 
 
 static int call_stream_address_gstring(GString *o, struct packet_stream *ps, enum stream_address_format format) {
@@ -507,6 +522,8 @@ INLINE char *bencode_get_alt(bencode_item_t *i, const char *one, const char *two
 }
 
 INLINE void ng_sdes_option(struct sdp_ng_flags *out, str *s, void *dummy) {
+	str_hyphenate(s);
+
 	if (call_ng_flags_prefix(out, s, "no-", call_ng_flags_str_ht, &out->sdes_no))
 		return;
 
@@ -547,6 +564,9 @@ INLINE void ng_sdes_option(struct sdp_ng_flags *out, str *s, void *dummy) {
 		case CSH_LOOKUP("pad"):
 			out->sdes_pad = 1;
 			break;
+		case CSH_LOOKUP("static"):
+			out->sdes_static = 1;
+			break;
 		default:
 			ilog(LOG_WARN, "Unknown 'SDES' flag encountered: '"STR_FORMAT"'",
 					STR_FMT(s));
@@ -692,37 +712,66 @@ static void call_ng_flags_rtcp_mux(struct sdp_ng_flags *out, str *s, void *dummy
 }
 static void call_ng_flags_replace(struct sdp_ng_flags *out, str *s, void *dummy) {
 	str_hyphenate(s);
-	if (!str_cmp(s, "origin"))
-		out->replace_origin = 1;
-	else if (!str_cmp(s, "session-connection"))
-		out->replace_sess_conn = 1;
-	else
-		ilog(LOG_WARN, "Unknown 'replace' flag encountered: '" STR_FORMAT "'",
-				STR_FMT(s));
+	switch (__csh_lookup(s)) {
+		case CSH_LOOKUP("origin"):
+			out->replace_origin = 1;
+			break;
+		case CSH_LOOKUP("session-connection"):
+			out->replace_sess_conn = 1;
+			break;
+		case CSH_LOOKUP("session-name"):
+			out->replace_sess_name = 1;
+			break;
+		case CSH_LOOKUP("sdp-version"):
+			out->replace_sdp_version = 1;
+			break;
+		case CSH_LOOKUP("SDP-version"):
+			out->replace_sdp_version = 1;
+			break;
+		case CSH_LOOKUP("username"):
+			out->replace_username = 1;
+			break;
+		case CSH_LOOKUP("zero-address"):
+			out->replace_zero_address = 1;
+			break;
+		default:
+			ilog(LOG_WARN, "Unknown 'replace' flag encountered: '" STR_FORMAT "'",
+					STR_FMT(s));
+	}
 }
 static void call_ng_flags_supports(struct sdp_ng_flags *out, str *s, void *dummy) {
 	if (!str_cmp(s, "load limit"))
 		out->supports_load_limit = 1;
 }
+static str *str_dup_escape(const str *s) {
+	str *ret = str_dup(s);
+	int i;
+	while ((i = str_str(ret, "--")) >= 0) {
+		ret->s[i] = '=';
+		memmove(&ret->s[i + 1], &ret->s[i + 2], ret->len - i - 2);
+		ret->len--;
+	}
+	return ret;
+}
 static void call_ng_flags_codec_list(struct sdp_ng_flags *out, str *s, void *qp) {
-	str *s_copy = str_slice_dup(s);
+	str *s_copy = str_dup_escape(s);
 	g_queue_push_tail((GQueue *) qp, s_copy);
 }
 static void call_ng_flags_str_ht(struct sdp_ng_flags *out, str *s, void *htp) {
-	str *s_copy = str_slice_dup(s);
+	str *s_copy = str_dup_escape(s);
 	GHashTable **ht = htp;
 	if (!*ht)
-		*ht = g_hash_table_new_full(str_case_hash, str_case_equal, str_slice_free, NULL);
+		*ht = g_hash_table_new_full(str_case_hash, str_case_equal, free, NULL);
 	g_hash_table_replace(*ht, s_copy, s_copy);
 }
 #ifdef WITH_TRANSCODING
 static void call_ng_flags_str_ht_split(struct sdp_ng_flags *out, str *s, void *htp) {
 	GHashTable **ht = htp;
 	if (!*ht)
-		*ht = g_hash_table_new_full(str_case_hash, str_case_equal, str_slice_free, str_slice_free);
+		*ht = g_hash_table_new_full(str_case_hash, str_case_equal, free, free);
 	str splitter = *s;
 	while (1) {
-		g_hash_table_replace(*ht, str_slice_dup(&splitter), str_slice_dup(s));
+		g_hash_table_replace(*ht, str_dup_escape(&splitter), str_dup_escape(s));
 		char *c = memrchr(splitter.s, '/', splitter.len);
 		if (!c)
 			break;
@@ -779,26 +828,39 @@ static void call_ng_flags_flags(struct sdp_ng_flags *out, str *s, void *dummy) {
 		case CSH_LOOKUP("port-latching"):
 			out->port_latching = 1;
 			break;
+		case CSH_LOOKUP("no-port-latching"):
+			out->no_port_latching = 1;
+			break;
 		case CSH_LOOKUP("generate-mid"):
 			out->generate_mid = 1;
 			break;
 		case CSH_LOOKUP("record-call"):
 			out->record_call = 1;
 			break;
+		case CSH_LOOKUP("debug"):
+			out->debug = 1;
+			break;
+		case CSH_LOOKUP("debugging"):
+			out->debug = 1;
+			break;
 		case CSH_LOOKUP("no-rtcp-attribute"):
 			out->no_rtcp_attr = 1;
 			break;
 		case CSH_LOOKUP("full-rtcp-attribute"):
 			out->full_rtcp_attr = 1;
 			break;
+		case CSH_LOOKUP("generate-RTCP"):
+			out->generate_rtcp = 1;
+			break;
 		case CSH_LOOKUP("loop-protect"):
 			out->loop_protect = 1;
 			break;
 		case CSH_LOOKUP("original-sendrecv"):
 			out->original_sendrecv = 1;
 			break;
-		case CSH_LOOKUP("always-transcode"):
-			out->always_transcode = 1;
+		case CSH_LOOKUP("always-transcode"):;
+			static const str str_all = STR_CONST_INIT("all");
+			call_ng_flags_str_ht(out, (str *) &str_all, &out->codec_accept);
 			break;
 		case CSH_LOOKUP("asymmetric-codecs"):
 			out->asymmetric_codecs = 1;
@@ -806,12 +868,30 @@ static void call_ng_flags_flags(struct sdp_ng_flags *out, str *s, void *dummy) {
 		case CSH_LOOKUP("symmetric-codecs"):
 			out->symmetric_codecs = 1;
 			break;
+		case CSH_LOOKUP("reorder-codecs"):
+			out->reorder_codecs = 1;
+			break;
+		case CSH_LOOKUP("single-codec"):
+			out->single_codec = 1;
+			break;
 		case CSH_LOOKUP("inject-DTMF"):
 			out->inject_dtmf = 1;
 			break;
 		case CSH_LOOKUP("pad-crypto"):
 			out->sdes_pad = 1;
 			break;
+		case CSH_LOOKUP("passthrough"):
+			out->passthrough_on = 1;
+			break;
+		case CSH_LOOKUP("no-passthrough"):
+			out->passthrough_off = 1;
+			break;
+		case CSH_LOOKUP("no-jitter-buffer"):
+			out->disable_jb = 1;
+			break;
+		case CSH_LOOKUP("pierce-NAT"):
+			out->pierce_nat = 1;
+			break;
 		default:
 			// handle values aliases from other dictionaries
 			if (call_ng_flags_prefix(out, s, "SDES-no-", call_ng_flags_str_ht, &out->sdes_no))
@@ -820,17 +900,17 @@ static void call_ng_flags_flags(struct sdp_ng_flags *out, str *s, void *dummy) {
 				return;
 			if (call_ng_flags_prefix(out, s, "OSRTP-", ng_osrtp_option, NULL))
 				return;
-			if (out->opmode == OP_OFFER) {
-				if (call_ng_flags_prefix(out, s, "codec-strip-", call_ng_flags_str_ht,
-							&out->codec_strip))
-					return;
-				if (call_ng_flags_prefix(out, s, "codec-offer-", call_ng_flags_codec_list,
-							&out->codec_offer))
-					return;
-				if (call_ng_flags_prefix(out, s, "codec-except-", call_ng_flags_str_ht,
-							&out->codec_except))
-					return;
+			if (call_ng_flags_prefix(out, s, "codec-strip-", call_ng_flags_str_ht,
+						&out->codec_strip))
+				return;
+			if (call_ng_flags_prefix(out, s, "codec-offer-", call_ng_flags_codec_list,
+						&out->codec_offer))
+				return;
+			if (call_ng_flags_prefix(out, s, "codec-except-", call_ng_flags_str_ht,
+						&out->codec_except))
+				return;
 #ifdef WITH_TRANSCODING
+			if (out->opmode == OP_OFFER) {
 				if (call_ng_flags_prefix(out, s, "transcode-", call_ng_flags_codec_list,
 							&out->codec_transcode))
 					return;
@@ -840,15 +920,21 @@ static void call_ng_flags_flags(struct sdp_ng_flags *out, str *s, void *dummy) {
 				if (call_ng_flags_prefix(out, s, "codec-mask-", call_ng_flags_str_ht,
 							&out->codec_mask))
 					return;
-				if (call_ng_flags_prefix(out, s, "codec-set-", call_ng_flags_str_ht_split,
-							&out->codec_set))
-					return;
 				if (call_ng_flags_prefix(out, s, "T38-", ng_t38_option, NULL))
 					return;
 				if (call_ng_flags_prefix(out, s, "T.38-", ng_t38_option, NULL))
 					return;
-#endif
 			}
+			if (call_ng_flags_prefix(out, s, "codec-set-", call_ng_flags_str_ht_split,
+						&out->codec_set))
+				return;
+			if (call_ng_flags_prefix(out, s, "codec-accept-", call_ng_flags_str_ht,
+						&out->codec_accept))
+				return;
+			if (call_ng_flags_prefix(out, s, "codec-consume-", call_ng_flags_str_ht,
+						&out->codec_consume))
+				return;
+#endif
 
 			ilog(LOG_WARN, "Unknown flag encountered: '" STR_FORMAT "'",
 					STR_FMT(s));
@@ -864,6 +950,7 @@ static void call_ng_process_flags(struct sdp_ng_flags *out, bencode_item_t *inpu
 
 	out->trust_address = trust_address_def;
 	out->dtls_passive = dtls_passive_def;
+	out->dtls_reverse_passive = dtls_passive_def;
 
 	call_ng_flags_list(out, input, "flags", call_ng_flags_flags, NULL);
 	call_ng_flags_list(out, input, "replace", call_ng_flags_replace, NULL);
@@ -907,15 +994,21 @@ static void call_ng_process_flags(struct sdp_ng_flags *out, bencode_item_t *inpu
 	if (bencode_dictionary_get_str(input, "ICE", &s)) {
 		switch (__csh_lookup(&s)) {
 			case CSH_LOOKUP("remove"):
-				out->ice_remove = 1;
+				out->ice_option = ICE_REMOVE;
 				break;
 			case CSH_LOOKUP("force"):
-				out->ice_force = 1;
+				out->ice_option = ICE_FORCE;
+				break;
+			case CSH_LOOKUP("default"):
+				out->ice_option = ICE_DEFAULT;
+				break;
+			case CSH_LOOKUP("optional"):
+				out->ice_option = ICE_OPTIONAL;
 				break;
 			case CSH_LOOKUP("force_relay"):
 			case CSH_LOOKUP("force-relay"):
 			case CSH_LOOKUP("force relay"):
-				out->ice_force_relay = 1;
+				out->ice_option = ICE_FORCE_RELAY;
 				break;
 			default:
 				ilog(LOG_WARN, "Unknown 'ICE' flag encountered: '"STR_FORMAT"'",
@@ -923,11 +1016,45 @@ static void call_ng_process_flags(struct sdp_ng_flags *out, bencode_item_t *inpu
 		}
 	}
 
+	if (bencode_dictionary_get_str(input, "ICE-lite", &s)) {
+		switch (__csh_lookup(&s)) {
+			case CSH_LOOKUP("off"):
+			case CSH_LOOKUP("none"):
+			case CSH_LOOKUP("no"):
+				out->ice_lite_option = ICE_LITE_OFF;
+				break;
+			case CSH_LOOKUP("forward"):
+			case CSH_LOOKUP("offer"):
+			case CSH_LOOKUP("fwd"):
+			case CSH_LOOKUP("fw"):
+				out->ice_lite_option = ICE_LITE_FWD;
+				break;
+			case CSH_LOOKUP("backward"):
+			case CSH_LOOKUP("backwards"):
+			case CSH_LOOKUP("reverse"):
+			case CSH_LOOKUP("answer"):
+			case CSH_LOOKUP("back"):
+			case CSH_LOOKUP("bkw"):
+			case CSH_LOOKUP("bk"):
+				out->ice_lite_option = ICE_LITE_BKW;
+				break;
+			case CSH_LOOKUP("both"):
+				out->ice_lite_option = ICE_LITE_BOTH;
+				break;
+			default:
+				ilog(LOG_WARN, "Unknown 'ICE-lite' flag encountered: '" STR_FORMAT "'",
+						STR_FMT(&s));
+		}
+	}
+
 	if (bencode_dictionary_get_str(input, "DTLS", &s)) {
 		switch (__csh_lookup(&s)) {
 			case CSH_LOOKUP("passive"):
 				out->dtls_passive = 1;
 				break;
+			case CSH_LOOKUP("active"):
+				out->dtls_passive = 0;
+				break;
 			case CSH_LOOKUP("no"):
 			case CSH_LOOKUP("off"):
 			case CSH_LOOKUP("disabled"):
@@ -940,6 +1067,40 @@ static void call_ng_process_flags(struct sdp_ng_flags *out, bencode_item_t *inpu
 		}
 	}
 
+	if (bencode_dictionary_get_str(input, "DTLS-reverse", &s)) {
+		switch (__csh_lookup(&s)) {
+			case CSH_LOOKUP("passive"):
+				out->dtls_reverse_passive = 1;
+				break;
+			case CSH_LOOKUP("active"):
+				out->dtls_reverse_passive = 0;
+				break;
+			default:
+				ilog(LOG_WARN, "Unknown 'DTLS-reverse' flag encountered: '"STR_FORMAT"'",
+						STR_FMT(&s));
+		}
+	}
+
+	if (bencode_dictionary_get_str(input, "passthrough", &s)) {
+		switch (__csh_lookup(&s)) {
+			case CSH_LOOKUP("on"):
+			case CSH_LOOKUP("yes"):
+			case CSH_LOOKUP("enable"):
+			case CSH_LOOKUP("enabled"):
+				out->passthrough_on = 1;
+				break;
+			case CSH_LOOKUP("no"):
+			case CSH_LOOKUP("off"):
+			case CSH_LOOKUP("disable"):
+			case CSH_LOOKUP("disabled"):
+				out->passthrough_off = 1;
+				break;
+			default:
+				ilog(LOG_WARN, "Unknown 'passthrough' flag encountered: '"STR_FORMAT"'",
+						STR_FMT(&s));
+		}
+	}
+
 	call_ng_flags_list(out, input, "rtcp-mux", call_ng_flags_rtcp_mux, NULL);
 	call_ng_flags_list(out, input, "SDES", ng_sdes_option, NULL);
 	call_ng_flags_list(out, input, "OSRTP", ng_osrtp_option, NULL);
@@ -961,6 +1122,7 @@ static void call_ng_process_flags(struct sdp_ng_flags *out, bencode_item_t *inpu
 	out->tos = bencode_dictionary_get_int_str(input, "TOS", 256);
 	bencode_get_alt(input, "record-call", "record call", &out->record_call_str);
 	bencode_dictionary_get_str(input, "metadata", &out->metadata);
+	bencode_dictionary_get_str(input, "DTLS-fingerprint", &out->dtls_fingerprint);
 
 	if (opmode == OP_OFFER) {
 		out->ptime = bencode_dictionary_get_int_str(input, "ptime", 0);
@@ -975,16 +1137,52 @@ static void call_ng_process_flags(struct sdp_ng_flags *out, bencode_item_t *inpu
 					STR_FMT(&s));
 	}
 
-	if (opmode == OP_OFFER && (dict = bencode_dictionary_get_expect(input, "codec", BENCODE_DICTIONARY))) {
+	if ((dict = bencode_dictionary_get_expect(input, "codec", BENCODE_DICTIONARY))) {
 		call_ng_flags_list(out, dict, "strip", call_ng_flags_str_ht, &out->codec_strip);
 		call_ng_flags_list(out, dict, "offer", call_ng_flags_codec_list, &out->codec_offer);
 		call_ng_flags_list(out, dict, "except", call_ng_flags_str_ht, &out->codec_except);
 #ifdef WITH_TRANSCODING
-		call_ng_flags_list(out, dict, "transcode", call_ng_flags_codec_list, &out->codec_transcode);
-		call_ng_flags_list(out, dict, "mask", call_ng_flags_str_ht, &out->codec_mask);
-		call_ng_flags_list(out, dict, "set", call_ng_flags_str_ht_split, &out->codec_set);
+		if (opmode == OP_OFFER) {
+			call_ng_flags_list(out, dict, "transcode", call_ng_flags_codec_list, &out->codec_transcode);
+			call_ng_flags_list(out, dict, "mask", call_ng_flags_str_ht, &out->codec_mask);
+			call_ng_flags_list(out, dict, "set", call_ng_flags_str_ht_split, &out->codec_set);
+			call_ng_flags_list(out, dict, "accept", call_ng_flags_str_ht, &out->codec_accept);
+			call_ng_flags_list(out, dict, "consume", call_ng_flags_str_ht, &out->codec_consume);
+		}
 #endif
 	}
+
+	if (bencode_get_alt(input, "generate-RTCP", "generate RTCP", &s)) {
+		if (!str_cmp(&s, "on"))
+			out->generate_rtcp = 1;
+		else if (!str_cmp(&s, "off"))
+			out->generate_rtcp_off = 1;
+	}
+
+	if (bencode_get_alt(input, "media-echo", "media echo", &s)) {
+		switch (__csh_lookup(&s)) {
+			case CSH_LOOKUP("blackhole"):
+			case CSH_LOOKUP("sinkhole"):
+				out->media_echo = MEO_BLACKHOLE;
+				break;
+			case CSH_LOOKUP("forward"):
+			case CSH_LOOKUP("fwd"):
+			case CSH_LOOKUP("fw"):
+				out->media_echo = MEO_FWD;
+				break;
+			case CSH_LOOKUP("backward"):
+			case CSH_LOOKUP("backwards"):
+			case CSH_LOOKUP("reverse"):
+			case CSH_LOOKUP("back"):
+			case CSH_LOOKUP("bkw"):
+			case CSH_LOOKUP("bk"):
+				out->media_echo = MEO_BKW;
+				break;
+			case CSH_LOOKUP("both"):
+				out->media_echo = MEO_BOTH;
+				break;
+		}
+	}
 }
 static void call_ng_free_flags(struct sdp_ng_flags *flags) {
 	if (flags->codec_strip)
@@ -995,10 +1193,14 @@ static void call_ng_free_flags(struct sdp_ng_flags *flags) {
 		g_hash_table_destroy(flags->codec_mask);
 	if (flags->codec_set)
 		g_hash_table_destroy(flags->codec_set);
+	if (flags->codec_accept)
+		g_hash_table_destroy(flags->codec_accept);
+	if (flags->codec_consume)
+		g_hash_table_destroy(flags->codec_consume);
 	if (flags->sdes_no)
 		g_hash_table_destroy(flags->sdes_no);
-	g_queue_clear_full(&flags->codec_offer, str_slice_free);
-	g_queue_clear_full(&flags->codec_transcode, str_slice_free);
+	g_queue_clear_full(&flags->codec_offer, free);
+	g_queue_clear_full(&flags->codec_transcode, free);
 }
 
 static enum load_limit_reasons call_offer_session_limit(void) {
@@ -1053,7 +1255,99 @@ static enum load_limit_reasons call_offer_session_limit(void) {
 	return ret;
 }
 
-static const char *call_offer_answer_ng(bencode_item_t *input,
+static void fragment_free(struct sdp_fragment *frag) {
+	streams_free(&frag->streams);
+	call_ng_free_flags(&frag->flags);
+	obj_put(frag->ngbuf);
+	g_slice_free1(sizeof(*frag), frag);
+}
+static void fragment_key_free(void *p) {
+	struct fragment_key *k = p;
+	free(k->call_id.s);
+	free(k->from_tag.s);
+	g_slice_free1(sizeof(*k), k);
+}
+static void queue_sdp_fragment(struct ng_buffer *ngbuf, GQueue *streams, struct sdp_ng_flags *flags) {
+	ilog(LOG_DEBUG, "Queuing up SDP fragment for " STR_FORMAT_M "/" STR_FORMAT_M,
+			STR_FMT_M(&flags->call_id), STR_FMT_M(&flags->from_tag));
+
+	struct fragment_key *k = g_slice_alloc0(sizeof(*k));
+	str_init_dup_str(&k->call_id, &flags->call_id);
+	str_init_dup_str(&k->from_tag, &flags->from_tag);
+
+	struct sdp_fragment *frag = g_slice_alloc0(sizeof(*frag));
+	frag->received = rtpe_now;
+	frag->ngbuf = obj_get(ngbuf);
+	frag->streams = *streams;
+	frag->flags = *flags;
+	g_queue_init(streams);
+	ZERO(*flags);
+
+	mutex_lock(&sdp_fragments_lock);
+	GQueue *frags = g_hash_table_lookup_queue_new(sdp_fragments, k, fragment_key_free);
+	g_queue_push_tail(frags, frag);
+	mutex_unlock(&sdp_fragments_lock);
+}
+#define MAX_FRAG_AGE 3000000
+static void dequeue_sdp_fragments(struct call_monologue *monologue) {
+	struct fragment_key k;
+	ZERO(k);
+	k.call_id = monologue->call->callid;
+	k.from_tag = monologue->tag;
+
+	mutex_lock(&sdp_fragments_lock);
+	GQueue *frags = g_hash_table_lookup(sdp_fragments, &k);
+	if (!frags) {
+		mutex_unlock(&sdp_fragments_lock);
+		return;
+	}
+
+	g_hash_table_remove(sdp_fragments, &k);
+	// we own the queue now
+	mutex_unlock(&sdp_fragments_lock);
+
+	struct sdp_fragment *frag;
+	while ((frag = g_queue_pop_head(frags))) {
+		if (timeval_diff(&rtpe_now, &frag->received) > MAX_FRAG_AGE)
+			goto next;
+
+		ilog(LOG_DEBUG, "Dequeuing SDP fragment for " STR_FORMAT_M "/" STR_FORMAT_M,
+				STR_FMT_M(&k.call_id), STR_FMT_M(&k.from_tag));
+
+		monologue_offer_answer(monologue, &frag->streams, &frag->flags);
+
+next:
+		fragment_free(frag);
+	}
+
+	g_queue_free(frags);
+}
+static gboolean fragment_check_cleanup(void *k, void *v, void *p) {
+	int all = GPOINTER_TO_INT(p);
+	struct fragment_key *key = k;
+	GQueue *frags = v;
+	if (!key || !frags)
+		return TRUE;
+	while (frags->length) {
+		struct sdp_fragment *frag = frags->head->data;
+		if (!all && timeval_diff(&rtpe_now, &frag->received) <= MAX_FRAG_AGE)
+			break;
+		g_queue_pop_head(frags);
+		fragment_free(frag);
+	}
+	if (!frags->length) {
+		g_queue_free(frags);
+		return TRUE;
+	}
+	return FALSE;
+}
+static void fragments_cleanup(int all) {
+	mutex_lock(&sdp_fragments_lock);
+	g_hash_table_foreach_remove(sdp_fragments, fragment_check_cleanup, GINT_TO_POINTER(all));
+	mutex_unlock(&sdp_fragments_lock);
+}
+
+static const char *call_offer_answer_ng(struct ng_buffer *ngbuf, bencode_item_t *input,
 		bencode_item_t *output, enum call_opmode opmode, const char* addr,
 		const endpoint_t *sin)
 {
@@ -1082,17 +1376,6 @@ static const char *call_offer_answer_ng(bencode_item_t *input,
 		str_swap(&flags.to_tag, &flags.from_tag);
 	}
 
-	if (opmode == OP_OFFER) {
-		enum load_limit_reasons limit = call_offer_session_limit();
-		if (limit != LOAD_LIMIT_NONE) {
-			if (!flags.supports_load_limit)
-				errstr = "Parallel session limit reached"; // legacy protocol
-			else
-				errstr = magic_load_limit_strings[limit];
-			goto out;
-		}
-	}
-
 	errstr = "Failed to parse SDP";
 	if (sdp_parse(&sdp, &parsed, &flags))
 		goto out;
@@ -1111,23 +1394,34 @@ static const char *call_offer_answer_ng(bencode_item_t *input,
 	/* OP_ANSWER; OP_OFFER && !IS_FOREIGN_CALL */
 	call = call_get(&flags.call_id);
 
-	/* Failover scenario because of timeout on offer response: siprouter tries
-	* to establish session with another rtpengine2 even though rtpengine1
-	* might have persisted part of the session. rtpengine2 deletes previous
-	* call in memory and recreates an OWN call in redis */
 	// SDP fragments for trickle ICE must always operate on an existing call
-	if (opmode == OP_OFFER && !flags.fragment) {
-		if (!call) {
-			/* call == NULL, should create call */
-			call = call_get_or_create(&flags.call_id, 0);
+	if (!call && opmode == OP_OFFER && flags.fragment) {
+		queue_sdp_fragment(ngbuf, &streams, &flags);
+		errstr = NULL;
+		goto out;
+	}
+
+	if (opmode == OP_OFFER && !call) {
+		enum load_limit_reasons limit = call_offer_session_limit();
+		if (limit != LOAD_LIMIT_NONE) {
+			if (!flags.supports_load_limit)
+				errstr = "Parallel session limit reached"; // legacy protocol
+			else
+				errstr = magic_load_limit_strings[limit];
+			goto out;
 		}
+
+		call = call_get_or_create(&flags.call_id, 0);
 	}
 
 	errstr = "Unknown call-id";
 	if (!call)
 		goto out;
 
-	if (IS_FOREIGN_CALL(call))
+	if (flags.debug)
+		call->debug = 1;
+
+	if (rtpe_config.active_switchover && IS_FOREIGN_CALL(call))
 		call_make_own_foreign(call, 0);
 
 	if (!call->created_from && addr) {
@@ -1173,12 +1467,21 @@ static const char *call_offer_answer_ng(bencode_item_t *input,
 		call->drop_traffic = 0;
 	}
 
+	int do_dequeue = 1;
+
 	ret = monologue_offer_answer(monologue, &streams, &flags);
 	if (!ret) {
 		// SDP fragments for trickle ICE are consumed with no replacement returned
 		if (!flags.fragment)
 			ret = sdp_replace(chopper, &parsed, monologue->active_dialogue, &flags);
 	}
+	else if (ret == ERROR_NO_ICE_AGENT && flags.fragment) {
+		queue_sdp_fragment(ngbuf, &streams, &flags);
+		ret = 0;
+		do_dequeue = 0;
+	}
+
+	// streams and flags are invalid after here
 
 	struct recording *recording = call->recording;
 	if (recording != NULL) {
@@ -1189,6 +1492,9 @@ static const char *call_offer_answer_ng(bencode_item_t *input,
 		recording_response(recording, output);
 	}
 
+	if (do_dequeue)
+		dequeue_sdp_fragments(monologue);
+
 	rwlock_unlock_w(&call->master_lock);
 
 	if (!flags.no_redis_update) {
@@ -1223,14 +1529,15 @@ out:
 	return errstr;
 }
 
-const char *call_offer_ng(bencode_item_t *input, bencode_item_t *output, const char* addr,
+const char *call_offer_ng(struct ng_buffer *ngbuf, bencode_item_t *input, bencode_item_t *output,
+		const char* addr,
 		const endpoint_t *sin)
 {
-	return call_offer_answer_ng(input, output, OP_OFFER, addr, sin);
+	return call_offer_answer_ng(ngbuf, input, output, OP_OFFER, addr, sin);
 }
 
-const char *call_answer_ng(bencode_item_t *input, bencode_item_t *output) {
-	return call_offer_answer_ng(input, output, OP_ANSWER, NULL, NULL);
+const char *call_answer_ng(struct ng_buffer *ngbuf, bencode_item_t *input, bencode_item_t *output) {
+	return call_offer_answer_ng(ngbuf, input, output, OP_ANSWER, NULL, NULL);
 }
 
 const char *call_delete_ng(bencode_item_t *input, bencode_item_t *output) {
@@ -1382,7 +1689,7 @@ static void ng_stats_media(bencode_item_t *list, const struct call_media *m,
 	BF_M("ICE", ICE);
 	BF_M("trickle ICE", TRICKLE_ICE);
 	BF_M("ICE controlling", ICE_CONTROLLING);
-	BF_M("ICE-lite", ICE_LITE);
+	BF_M("ICE-lite peer", ICE_LITE_PEER);
 	BF_M("unidirectional", UNIDIRECTIONAL);
 	BF_M("loop check", LOOP_CHECK);
 	BF_M("transcoding", TRANSCODE);
@@ -1396,7 +1703,7 @@ stats:
 }
 
 static void ng_stats_monologue(bencode_item_t *dict, const struct call_monologue *ml,
-		struct call_stats *totals)
+		struct call_stats *totals, bencode_item_t *ssrc)
 {
 	bencode_item_t *sub, *medias = NULL;
 	GList *l;
@@ -1418,6 +1725,8 @@ static void ng_stats_monologue(bencode_item_t *dict, const struct call_monologue
 	bencode_dictionary_add_integer(sub, "created", ml->created);
 	if (ml->active_dialogue)
 		bencode_dictionary_add_str(sub, "in dialogue with", &ml->active_dialogue->tag);
+	ng_stats_ssrc(bencode_dictionary_add_dictionary(sub, "SSRC"), ml->ssrc_hash);
+	ng_stats_ssrc(ssrc, ml->ssrc_hash);
 
 	medias = bencode_dictionary_add_list(sub, "medias");
 
@@ -1433,6 +1742,7 @@ static void ng_stats_ssrc_mos_entry_common(bencode_item_t *subent, struct ssrc_s
 {
 	bencode_dictionary_add_integer(subent, "MOS", sb->mos / div);
 	bencode_dictionary_add_integer(subent, "round-trip time", sb->rtt / div);
+	bencode_dictionary_add_integer(subent, "round-trip time leg", sb->rtt_leg / div);
 	bencode_dictionary_add_integer(subent, "jitter", sb->jitter / div);
 	bencode_dictionary_add_integer(subent, "packet loss", sb->packetloss / div);
 }
@@ -1459,12 +1769,18 @@ static void ng_stats_ssrc(bencode_item_t *dict, struct ssrc_hash *ht) {
 		struct ssrc_entry_call *se = l->data;
 		char *tmp = bencode_buffer_alloc(dict->buffer, 12);
 		snprintf(tmp, 12, "%" PRIu32, se->h.ssrc);
+		if (bencode_dictionary_get(dict, tmp))
+			continue;
 		bencode_item_t *ent = bencode_dictionary_add_dictionary(dict, tmp);
 
 		if (!se->stats_blocks.length || !se->lowest_mos || !se->highest_mos)
 			continue;
 
-		ng_stats_ssrc_mos_entry_dict_avg(ent, "average MOS", &se->average_mos, se->stats_blocks.length);
+		bencode_dictionary_add_integer(ent, "cumulative loss", se->packets_lost);
+
+		int mos_samples = se->stats_blocks.length - se->no_mos_count;
+		if (mos_samples < 1) mos_samples = 1;
+		ng_stats_ssrc_mos_entry_dict_avg(ent, "average MOS", &se->average_mos, mos_samples);
 		ng_stats_ssrc_mos_entry_dict(ent, "lowest MOS", se->lowest_mos);
 		ng_stats_ssrc_mos_entry_dict(ent, "highest MOS", se->highest_mos);
 
@@ -1502,6 +1818,7 @@ void ng_call_stats(struct call *call, const str *fromtag, const str *totag, benc
 	GList *l;
 	struct call_monologue *ml;
 	struct call_stats t_b;
+	bencode_item_t *ssrc = NULL;
 
 	if (!totals)
 		totals = &t_b;
@@ -1515,7 +1832,7 @@ void ng_call_stats(struct call *call, const str *fromtag, const str *totag, benc
 	bencode_dictionary_add_integer(output, "created", call->created.tv_sec);
 	bencode_dictionary_add_integer(output, "created_us", call->created.tv_usec);
 	bencode_dictionary_add_integer(output, "last signal", call->last_signal);
-	ng_stats_ssrc(bencode_dictionary_add_dictionary(output, "SSRC"), call->ssrc_hash);
+	ssrc = bencode_dictionary_add_dictionary(output, "SSRC");
 
 	tags = bencode_dictionary_add_dictionary(output, "tags");
 
@@ -1525,14 +1842,14 @@ stats:
 	if (!match_tag || !match_tag->len) {
 		for (l = call->monologues.head; l; l = l->next) {
 			ml = l->data;
-			ng_stats_monologue(tags, ml, totals);
+			ng_stats_monologue(tags, ml, totals, ssrc);
 		}
 	}
 	else {
 		ml = g_hash_table_lookup(call->tags, match_tag);
 		if (ml) {
-			ng_stats_monologue(tags, ml, totals);
-			ng_stats_monologue(tags, ml->active_dialogue, totals);
+			ng_stats_monologue(tags, ml, totals, ssrc);
+			ng_stats_monologue(tags, ml->active_dialogue, totals, ssrc);
 		}
 	}
 
@@ -1621,15 +1938,17 @@ const char *call_start_recording_ng(bencode_item_t *input, bencode_item_t *outpu
 const char *call_stop_recording_ng(bencode_item_t *input, bencode_item_t *output) {
 	str callid;
 	struct call *call;
+	str metadata;
 
 	if (!bencode_dictionary_get_str(input, "call-id", &callid))
 		return "No call-id in message";
+	bencode_dictionary_get_str(input, "metadata", &metadata);
 	call = call_get_opmode(&callid, OP_OTHER);
 	if (!call)
 		return "Unknown call-id";
 
 	call->recording_on = 0;
-	recording_stop(call);
+	recording_stop(call, &metadata);
 
 	rwlock_unlock_w(&call->master_lock);
 	obj_put(call);
@@ -1755,7 +2074,7 @@ const char *call_stop_forwarding_ng(bencode_item_t *input, bencode_item_t *outpu
 		}
 	}
 
-	recording_stop(call);
+	recording_stop(call, NULL);
 
 	errstr = NULL;
 out:
@@ -1934,6 +2253,7 @@ const char *call_play_media_ng(bencode_item_t *input, bencode_item_t *output) {
 	GQueue monologues;
 	const char *err = NULL;
 	long long db_id;
+	long long repeat_times = 1;
 
 	err = play_media_select_party(&call, &monologues, input);
 	if (err)
@@ -1944,21 +2264,21 @@ const char *call_play_media_ng(bencode_item_t *input, bencode_item_t *output) {
 
 		if (!monologue->player)
 			monologue->player = media_player_new(monologue);
-
+		repeat_times = bencode_dictionary_get_int_str(input, "repeat-times", 1);
 		err = "No media file specified";
 		if (bencode_dictionary_get_str(input, "file", &str)) {
 			err = "Failed to start media playback from file";
-			if (media_player_play_file(monologue->player, &str))
+			if (media_player_play_file(monologue->player, &str,repeat_times))
 				goto out;
 		}
 		else if (bencode_dictionary_get_str(input, "blob", &str)) {
 			err = "Failed to start media playback from blob";
-			if (media_player_play_blob(monologue->player, &str))
+			if (media_player_play_blob(monologue->player, &str,repeat_times))
 				goto out;
 		}
 		else if ((db_id = bencode_dictionary_get_int_str(input, "db-id", 0)) > 0) {
 			err = "Failed to start media playback from database";
-			if (media_player_play_db(monologue->player, db_id))
+			if (media_player_play_db(monologue->player, db_id,repeat_times))
 				goto out;
 		}
 		else
@@ -2131,6 +2451,26 @@ void call_interfaces_free() {
 		pcre_free_study(streams_ree);
 		streams_ree = NULL;
 	}
+
+	fragments_cleanup(1);
+	g_hash_table_destroy(sdp_fragments);
+	sdp_fragments = NULL;
+	mutex_destroy(&sdp_fragments_lock);
+}
+
+void call_interfaces_timer() {
+	fragments_cleanup(0);
+}
+
+static unsigned int frag_key_hash(const void *A) {
+	const struct fragment_key *a = A;
+	return str_hash(&a->call_id) ^ str_hash(&a->from_tag);
+}
+static int frag_key_eq(const void *A, const void *B) {
+	const struct fragment_key *a = A;
+	const struct fragment_key *b = B;
+	return str_equal(&a->call_id, &b->call_id)
+		&& str_equal(&a->from_tag, &b->from_tag);
 }
 
 int call_interfaces_init() {
@@ -2147,5 +2487,8 @@ int call_interfaces_init() {
 		return -1;
 	streams_ree = pcre_study(streams_re, 0, &errptr);
 
+	sdp_fragments = g_hash_table_new_full(frag_key_hash, frag_key_eq, fragment_key_free, NULL);
+	mutex_init(&sdp_fragments_lock);
+
 	return 0;
 }
