@@ -6,6 +6,7 @@
 #include <sys/resource.h>
 #include <glib.h>
 #include <glib/gstdio.h>
+#include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -14,6 +15,7 @@
 #include <time.h>
 #include <ifaddrs.h>
 #include <net/if.h>
+#include <netdb.h>
 
 #include "poller.h"
 #include "control_tcp.h"
@@ -45,10 +47,13 @@
 #include "media_player.h"
 #include "dtmf.h"
 #include "jitter_buffer.h"
+#include "websocket.h"
+#include "codec.h"
 
 
 
 struct poller *rtpe_poller;
+struct poller_map *rtpe_poller_map;
 struct rtpengine_config initial_rtpe_config;
 
 static struct control_tcp *rtpe_tcp;
@@ -74,7 +79,17 @@ struct rtpengine_config rtpe_config = {
 	.redis_connect_timeout = 1000,
 	.media_num_threads = -1,
 	.dtls_rsa_key_size = 2048,
+	.dtls_mtu = 1200, // chrome default mtu
 	.dtls_signature = 256,
+	.max_dtx = 30,
+	.dtx_shift = 5,
+	.dtx_buffer = 10,
+	.dtx_lag = 100,
+	.common = {
+		.log_levels = {
+			[log_level_index_internals] = -1,
+		},
+	},
 };
 
 static void sighandler(gpointer x) {
@@ -102,17 +117,19 @@ static void sighandler(gpointer x) {
 		if (ret == SIGINT || ret == SIGTERM)
 			rtpe_shutdown = 1;
 		else if (ret == SIGUSR1) {
-		        if (get_log_level() > 0) {
-				g_atomic_int_add(&rtpe_config.common.log_level, -1);
-				ilog(get_log_level(), "Set log level to %d\n",
-						get_log_level());
+			for (unsigned int i = 0; i < num_log_levels; i++) {
+				g_atomic_int_add(&rtpe_config.common.log_levels[i], -1);
+				ilogsn(i, __get_log_level(i), "Decreased '%s' log level to %d\n",
+						log_level_names[i],
+						__get_log_level(i));
 			}
 		}
 		else if (ret == SIGUSR2) {
-		        if (get_log_level() < 7) {
-				g_atomic_int_add(&rtpe_config.common.log_level, 1);
-				ilog(get_log_level(), "Set log level to %d\n",
-						get_log_level());
+			for (unsigned int i = 0; i < num_log_levels; i++) {
+				g_atomic_int_add(&rtpe_config.common.log_levels[i], 1);
+				ilogsn(i, __get_log_level(i), "Increased '%s' log level to %d\n",
+						log_level_names[i],
+						__get_log_level(i));
 			}
 		}
 		else
@@ -150,6 +167,86 @@ static void resources(void) {
 
 
 
+static void __find_if_name(char *s, struct ifaddrs *ifas, GQueue *addrs) {
+	sockaddr_t *addr;
+
+	for (struct ifaddrs *ifa = ifas; ifa; ifa = ifa->ifa_next) {
+		if (strcmp(ifa->ifa_name, s))
+			continue;
+		if (!(ifa->ifa_flags & IFF_UP))
+			continue;
+		if (!ifa->ifa_addr)
+			continue;
+
+		addr = g_slice_alloc(sizeof(*addr));
+		if (ifa->ifa_addr->sa_family == AF_INET) {
+			struct sockaddr_in *sin = (void *) ifa->ifa_addr;
+			addr->family = __get_socket_family_enum(SF_IP4);
+			addr->u.ipv4 = sin->sin_addr;
+		}
+		else if (ifa->ifa_addr->sa_family == AF_INET6) {
+			struct sockaddr_in6 *sin = (void *) ifa->ifa_addr;
+			if (sin->sin6_scope_id) {
+				// link-local
+				g_slice_free1(sizeof(*addr), addr);
+				continue;
+			}
+			addr->family = __get_socket_family_enum(SF_IP6);
+			addr->u.ipv6 = sin->sin6_addr;
+		}
+		else {
+			g_slice_free1(sizeof(*addr), addr);
+			continue;
+		}
+
+		// got one
+		ilog(LOG_DEBUG, "Determined address %s for interface '%s'",
+				sockaddr_print_buf(addr), s);
+		g_queue_push_tail(addrs, addr);
+	}
+}
+
+static void __resolve_ifname(char *s, GQueue *addrs) {
+	struct addrinfo hints = {
+		.ai_family = AF_UNSPEC,
+		.ai_socktype = SOCK_DGRAM,
+	};
+
+	struct addrinfo *res = NULL;
+	int status = getaddrinfo(s, NULL, &hints, &res);
+	if (status) {
+		ilog(LOG_ERR, "Failed to resolve '%s' as a DNS host name: %s", s, gai_strerror(status));
+		return;
+	}
+
+	for (struct addrinfo *r = res; r; r = r->ai_next) {
+		sockaddr_t *addr = g_slice_alloc0(sizeof(*addr));
+
+		if (r->ai_family == AF_INET) {
+			struct sockaddr_in *sin = (void *) r->ai_addr;
+			assert(r->ai_addrlen >= sizeof(*sin));
+			addr->family = __get_socket_family_enum(SF_IP4);
+			addr->u.ipv4 = sin->sin_addr;
+		}
+		else if (r->ai_family == AF_INET6) {
+			struct sockaddr_in6 *sin = (void *) r->ai_addr;
+			assert(r->ai_addrlen >= sizeof(*sin));
+			addr->family = __get_socket_family_enum(SF_IP6);
+			addr->u.ipv6 = sin->sin6_addr;
+		}
+		else {
+			g_slice_free1(sizeof(*addr), addr);
+			continue;
+		}
+
+		ilog(LOG_DEBUG, "Determined address %s for host name '%s'",
+				sockaddr_print_buf(addr), s);
+		g_queue_push_tail(addrs, addr);
+	}
+
+	freeaddrinfo(res);
+}
+
 static int if_addr_parse(GQueue *q, char *s, struct ifaddrs *ifas) {
 	str name;
 	char *c;
@@ -180,40 +277,16 @@ static int if_addr_parse(GQueue *q, char *s, struct ifaddrs *ifas) {
 		g_queue_push_tail(&addrs, addr);
 	}
 	else {
+		g_slice_free1(sizeof(*addr), addr);
 		// could be an interface name?
 		ilog(LOG_DEBUG, "Could not parse '%s' as network address, checking to see if "
 				"it's an interface", s);
-		for (struct ifaddrs *ifa = ifas; ifa; ifa = ifa->ifa_next) {
-			if (strcmp(ifa->ifa_name, s))
-				continue;
-			if (!(ifa->ifa_flags & IFF_UP))
-				continue;
-			if (!ifa->ifa_addr)
-				continue;
-			if (ifa->ifa_addr->sa_family == AF_INET) {
-				struct sockaddr_in *sin = (void *) ifa->ifa_addr;
-				addr->family = __get_socket_family_enum(SF_IP4);
-				addr->u.ipv4 = sin->sin_addr;
-			}
-			else if (ifa->ifa_addr->sa_family == AF_INET6) {
-				struct sockaddr_in6 *sin = (void *) ifa->ifa_addr;
-				if (sin->sin6_scope_id)
-					continue; // link-local
-				addr->family = __get_socket_family_enum(SF_IP6);
-				addr->u.ipv6 = sin->sin6_addr;
-			}
-			else
-				continue;
+		__find_if_name(s, ifas, &addrs);
 
-			// got one
-			ilog(LOG_DEBUG, "Determined address %s for interface '%s'",
-					sockaddr_print_buf(addr), s);
-			g_queue_push_tail(&addrs, addr);
-			addr = g_slice_alloc(sizeof(*addr));
+		if (!addrs.length) {
+			ilog(LOG_DEBUG, "'%s' is not an interface, attempting to resolve it as DNS host name", s);
+			__resolve_ifname(s, &addrs);
 		}
-
-		// free last unused entry
-		g_slice_free1(sizeof(*addr), addr);
 	}
 
 	if (!addrs.length) // nothing found
@@ -221,8 +294,13 @@ static int if_addr_parse(GQueue *q, char *s, struct ifaddrs *ifas) {
 
 	ZERO(adv);
 	if (c) {
-		if (sockaddr_parse_any(&adv, c))
-			return -1;
+		if (sockaddr_parse_any(&adv, c)) {
+			ilog(LOG_DEBUG, "Could not parse '%s' as an address, attempting DNS lookup", c);
+			if (sockaddr_getaddrinfo(&adv, c)) {
+				ilog(LOG_WARN, "DNS lookup for '%s' failed", c);
+				return -1;
+			}
+		}
 		if (is_addr_unspecified(&adv))
 			return -1;
 	}
@@ -257,7 +335,7 @@ static int redis_ep_parse(endpoint_t *ep, int *db, char **auth, const char *auth
 	char *sl;
 	long l;
 
-	sl = strchr(str, '@');
+	sl = strrchr(str, '@');
 	if (sl) {
 		*sl = 0;
 		*auth = g_strdup(str);
@@ -285,6 +363,32 @@ static int redis_ep_parse(endpoint_t *ep, int *db, char **auth, const char *auth
 }
 
 
+static void parse_cn_payload(str *out, char **in, const char *def, const char *name) {
+	if (!in || !*in) {
+		if (def)
+			str_init_dup(out, def);
+		return;
+	}
+
+	int len = g_strv_length(in);
+	if (len < 1)
+		die("Invalid CN payload specified (--%s)", name);
+	out->s = malloc(len);
+	for (int i = 0; i < len; i++) {
+		char *endp;
+		long p = strtol(in[i], &endp, 0);
+		if (endp == in[i] || *endp != '\0')
+			die("Invalid CN payload specified (--%s)", name);
+		if (p < 0 || p > 254)
+			die("Invalid CN payload specified (--%s)", name);
+		if (i == 0 && p > 127)
+			die("Invalid CN payload specified (--%s)", name);
+		out->s[i] = p;
+	}
+	out->len = len;
+}
+
+
 
 static void options(int *argc, char ***argv) {
 	AUTO_CLEANUP_GVBUF(if_a);
@@ -295,6 +399,7 @@ static void options(int *argc, char ***argv) {
 	AUTO_CLEANUP_GBUF(listenps);
 	AUTO_CLEANUP_GBUF(listenudps);
 	AUTO_CLEANUP_GBUF(listenngs);
+	AUTO_CLEANUP_GBUF(listenngtcps);
 	AUTO_CLEANUP_GBUF(listencli);
 	AUTO_CLEANUP_GBUF(graphitep);
 	AUTO_CLEANUP_GBUF(graphite_prefix_s);
@@ -314,15 +419,24 @@ static void options(int *argc, char ***argv) {
 	AUTO_CLEANUP_GBUF(dtmf_udp_ep);
 	AUTO_CLEANUP_GBUF(endpoint_learning);
 	AUTO_CLEANUP_GBUF(dtls_sig);
+	double silence_detect = 0;
+	AUTO_CLEANUP_GVBUF(cn_payload);
+	AUTO_CLEANUP_GVBUF(dtx_cn_params);
+	int debug_srtp = 0;
+	AUTO_CLEANUP_GBUF(amr_dtx);
+
+	rwlock_lock_w(&rtpe_config.config_lock);
 
 	GOptionEntry e[] = {
 		{ "table",	't', 0, G_OPTION_ARG_INT,	&rtpe_config.kernel_table,		"Kernel table to use",		"INT"		},
 		{ "no-fallback",'F', 0, G_OPTION_ARG_NONE,	&rtpe_config.no_fallback,	"Only start when kernel module is available", NULL },
 		{ "interface",	'i', 0, G_OPTION_ARG_STRING_ARRAY,&if_a,	"Local interface for RTP",	"[NAME/]IP[!IP]"},
+		{ "save-interface-ports",'S', 0, G_OPTION_ARG_NONE,	&rtpe_config.save_interface_ports,	"Bind ports only on first available interface of desired family", NULL },
 		{ "subscribe-keyspace", 'k', 0, G_OPTION_ARG_STRING_ARRAY,&ks_a,	"Subscription keyspace list",	"INT INT ..."},
 		{ "listen-tcp",	'l', 0, G_OPTION_ARG_STRING,	&listenps,	"TCP port to listen on",	"[IP:]PORT"	},
 		{ "listen-udp",	'u', 0, G_OPTION_ARG_STRING,	&listenudps,	"UDP port to listen on",	"[IP46|HOSTNAME:]PORT"	},
 		{ "listen-ng",	'n', 0, G_OPTION_ARG_STRING,	&listenngs,	"UDP port to listen on, NG protocol", "[IP46|HOSTNAME:]PORT"	},
+		{ "listen-tcp-ng",	'N', 0, G_OPTION_ARG_STRING,	&listenngtcps,	"TCP port to listen on, NG protocol", "[IP46|HOSTNAME:]PORT"	},
 		{ "listen-cli", 'c', 0, G_OPTION_ARG_STRING,    &listencli,     "UDP port to listen on, CLI",   "[IP46|HOSTNAME:]PORT"     },
 		{ "graphite", 'g', 0, G_OPTION_ARG_STRING,    &graphitep,     "Address of the graphite server",   "IP46|HOSTNAME:PORT"     },
 		{ "graphite-interval",  'G', 0, G_OPTION_ARG_INT,    &rtpe_config.graphite_interval,  "Graphite send interval in seconds",    "INT"   },
@@ -344,12 +458,19 @@ static void options(int *argc, char ***argv) {
 		{ "redis-disable-time", 0, 0, G_OPTION_ARG_INT, &rtpe_config.redis_disable_time, "Number of seconds redis communication is disabled because of errors", "INT" },
 		{ "redis-cmd-timeout", 0, 0, G_OPTION_ARG_INT, &rtpe_config.redis_cmd_timeout, "Sets a timeout in milliseconds for redis commands", "INT" },
 		{ "redis-connect-timeout", 0, 0, G_OPTION_ARG_INT, &rtpe_config.redis_connect_timeout, "Sets a timeout in milliseconds for redis connections", "INT" },
+		{ "redis-delete-async", 'y', 0, G_OPTION_ARG_INT, &rtpe_config.redis_delete_async, "Enable asynchronous redis delete", NULL },
+		{ "redis-delete-async-interval", 'y', 0, G_OPTION_ARG_INT, &rtpe_config.redis_delete_async_interval, "Set asynchronous redis delete interval (seconds)", NULL },
+		{ "active-switchover", 0,0,G_OPTION_ARG_NONE,	&rtpe_config.active_switchover, "Use call activity as indicator of active/standby state", NULL },
 		{ "b2b-url",	'b', 0, G_OPTION_ARG_STRING,	&rtpe_config.b2b_url,	"XMLRPC URL of B2B UA"	,	"STRING"	},
 		{ "log-facility-cdr",0,  0, G_OPTION_ARG_STRING, &log_facility_cdr_s, "Syslog facility to use for logging CDRs", "daemon|local0|...|local7"},
 		{ "log-facility-rtcp",0,  0, G_OPTION_ARG_STRING, &log_facility_rtcp_s, "Syslog facility to use for logging RTCP", "daemon|local0|...|local7"},
+#ifdef WITH_TRANSCODING
 		{ "log-facility-dtmf",0,  0, G_OPTION_ARG_STRING, &log_facility_dtmf_s, "Syslog facility to use for logging DTMF", "daemon|local0|...|local7"},
-		{ "log-format",	0, 0,	G_OPTION_ARG_STRING,	&log_format,	"Log prefix format",		"default|parsable"},
 		{ "dtmf-log-dest", 0,0,	G_OPTION_ARG_STRING,	&dtmf_udp_ep,	"Destination address for DTMF logging via UDP",	"IP46|HOSTNAME:PORT"	},
+		{ "dtmf-log-ng-tcp", 0,0,	G_OPTION_ARG_NONE,	&rtpe_config.dtmf_via_ng,	"DTMF logging via TCP NG protocol",	NULL },
+		{ "dtmf-no-suppress", 0,0,G_OPTION_ARG_NONE,	&rtpe_config.dtmf_no_suppress,	"Disable audio suppression during DTMF events",	NULL },
+#endif
+		{ "log-format",	0, 0,	G_OPTION_ARG_STRING,	&log_format,	"Log prefix format",		"default|parsable"},
 		{ "xmlrpc-format",'x', 0, G_OPTION_ARG_INT,	&rtpe_config.fmt,	"XMLRPC timeout request format to use. 0: SEMS DI, 1: call-id only, 2: Kamailio",	"INT"	},
 		{ "num-threads",  0, 0, G_OPTION_ARG_INT,	&rtpe_config.num_threads,	"Number of worker threads to create",	"INT"	},
 		{ "media-num-threads",  0, 0, G_OPTION_ARG_INT,	&rtpe_config.media_num_threads,	"Number of worker threads for media playback",	"INT"	},
@@ -383,10 +504,30 @@ static void options(int *argc, char ***argv) {
 		{ "endpoint-learning",0,0,G_OPTION_ARG_STRING,	&endpoint_learning,	"RTP endpoint learning algorithm",	"delayed|immediate|off|heuristic"	},
 		{ "jitter-buffer",0, 0,	G_OPTION_ARG_INT,	&rtpe_config.jb_length,	"Size of jitter buffer",		"INT" },
 		{ "jb-clock-drift",0,0,	G_OPTION_ARG_NONE,	&rtpe_config.jb_clock_drift,"Compensate for source clock drift",NULL },
-		{ "debug-srtp",0,0,	G_OPTION_ARG_NONE,	&rtpe_config.debug_srtp,"Log raw encryption details for SRTP",	NULL },
+		{ "debug-srtp",0,0,	G_OPTION_ARG_NONE,	&debug_srtp,		"Log raw encryption details for SRTP",	NULL },
 		{ "dtls-rsa-key-size",0, 0,	G_OPTION_ARG_INT,&rtpe_config.dtls_rsa_key_size,"Size of RSA key for DTLS",	"INT"		},
+		{ "dtls-mtu",0, 0,	G_OPTION_ARG_INT,&rtpe_config.dtls_mtu,"DTLS MTU",	"INT"		},
 		{ "dtls-ciphers",0,  0,	G_OPTION_ARG_STRING,	&rtpe_config.dtls_ciphers,"List of ciphers for DTLS",		"STRING"	},
 		{ "dtls-signature",0,  0,G_OPTION_ARG_STRING,	&dtls_sig,		"Signature algorithm for DTLS",		"SHA-256|SHA-1"	},
+		{ "listen-http", 0,0,	G_OPTION_ARG_STRING_ARRAY,&rtpe_config.http_ifs,"Interface for HTTP and WS",	"[IP46|HOSTNAME:]PORT"},
+		{ "listen-https", 0,0,	G_OPTION_ARG_STRING_ARRAY,&rtpe_config.https_ifs,"Interface for HTTPS and WSS",	"[IP46|HOSTNAME:]PORT"},
+		{ "https-cert", 0,0,	G_OPTION_ARG_STRING,	&rtpe_config.https_cert,"Certificate for HTTPS and WSS","FILE"},
+		{ "https-key", 0,0,	G_OPTION_ARG_STRING,	&rtpe_config.https_key,	"Private key for HTTPS and WSS","FILE"},
+		{ "http-threads", 0,0,	G_OPTION_ARG_INT,	&rtpe_config.http_threads,"Number of worker threads for HTTP and WS","INT"},
+		{ "software-id", 0,0,	G_OPTION_ARG_STRING,	&rtpe_config.software_id,"Identification string of this software presented to external systems","STRING"},
+		{ "poller-per-thread", 0,0,	G_OPTION_ARG_NONE,	&rtpe_config.poller_per_thread,	"Use poller per thread",	NULL },
+#ifdef WITH_TRANSCODING
+		{ "dtx-delay",	0,0,	G_OPTION_ARG_INT,	&rtpe_config.dtx_delay,	"Delay in milliseconds to trigger DTX handling","INT"},
+		{ "max-dtx",	0,0,	G_OPTION_ARG_INT,	&rtpe_config.max_dtx,	"Maximum duration of DTX handling",	"INT"},
+		{ "dtx-buffer",	0,0,	G_OPTION_ARG_INT,	&rtpe_config.dtx_buffer,"Maxmium number of packets held in DTX buffer",	"INT"},
+		{ "dtx-lag",	0,0,	G_OPTION_ARG_INT,	&rtpe_config.dtx_lag,	"Maxmium time span in milliseconds held in DTX buffer",	"INT"},
+		{ "dtx-shift",	0,0,	G_OPTION_ARG_INT,	&rtpe_config.dtx_shift,	"Length of time (in ms) to shift DTX buffer after over/underflow",	"INT"},
+		{ "dtx-cn-params",0,0,	G_OPTION_ARG_STRING_ARRAY,&dtx_cn_params,	"Parameters for CN generated from DTX","INT INT INT ..."},
+		{ "amr-dtx", 0,0,	G_OPTION_ARG_STRING,	&amr_dtx,		"DTX mechanism to use for AMR and AMR-WB","native|CN"},
+		{ "silence-detect",0,0,	G_OPTION_ARG_DOUBLE,	&silence_detect,	"Audio level threshold in percent for silence detection","FLOAT"},
+		{ "cn-payload",0,0,	G_OPTION_ARG_STRING_ARRAY,&cn_payload,		"Comfort noise parameters to replace silence with","INT INT INT ..."},
+		{ "reorder-codecs",0,0,	G_OPTION_ARG_NONE,	&rtpe_config.reorder_codecs,"Reorder answer codecs based on sender preference",NULL},
+#endif
 
 		{ NULL, }
 	};
@@ -411,8 +552,11 @@ static void options(int *argc, char ***argv) {
 
 	if (!if_a)
 		die("Missing option --interface");
-	if (!listenps && !listenudps && !listenngs)
-		die("Missing option --listen-tcp, --listen-udp or --listen-ng");
+	if (!listenps && !listenudps && !listenngs && !listenngtcps
+			&& !(rtpe_config.http_ifs && rtpe_config.http_ifs[0])
+			&& !(rtpe_config.https_ifs && rtpe_config.https_ifs[0]))
+		die("Missing option --listen-tcp, --listen-udp, --listen-ng, --listen-tcp-ng, "
+				"--listen-http, or --listen-https");
 
 	struct ifaddrs *ifas;
 	if (getifaddrs(&ifas)) {
@@ -438,9 +582,9 @@ static void options(int *argc, char ***argv) {
 
 			if ((errno == ERANGE && (uint_keyspace_db == ULONG_MAX)) ||
 			    (errno != 0 && uint_keyspace_db == 0)) {
-				ilog(LOG_ERR, "Fail adding keyspace '%.*s' to redis notifications; errono=%d\n", str_keyspace_db.len, str_keyspace_db.s, errno);
+				ilog(LOG_ERR, "Fail adding keyspace '" STR_FORMAT "' to redis notifications; errono=%d\n", STR_FMT(&str_keyspace_db), errno);
 			} else if (endptr == str_keyspace_db.s) {
-				ilog(LOG_ERR, "Fail adding keyspace '%.*s' to redis notifications; no digits found\n", str_keyspace_db.len, str_keyspace_db.s);
+				ilog(LOG_ERR, "Fail adding keyspace '" STR_FORMAT "' to redis notifications; no digits found\n", STR_FMT(&str_keyspace_db));
 			} else {
 				g_queue_push_tail(&rtpe_config.redis_subscribed_keyspaces, GUINT_TO_POINTER(uint_keyspace_db));
 			}
@@ -459,6 +603,10 @@ static void options(int *argc, char ***argv) {
 		if (endpoint_parse_any_getaddrinfo(&rtpe_config.ng_listen_ep, listenngs))
 			die("Invalid IP or port '%s' (--listen-ng)", listenngs);
 	}
+	if (listenngtcps) {
+		if (endpoint_parse_any_getaddrinfo(&rtpe_config.ng_tcp_listen_ep, listenngtcps))
+			die("Invalid IP or port '%s' (--listen-tcp-ng)", listenngtcps);
+	}
 
 	if (listencli) {if (endpoint_parse_any_getaddrinfo(&rtpe_config.cli_listen_ep, listencli))
 	    die("Invalid IP or port '%s' (--listen-cli)", listencli);
@@ -545,6 +693,9 @@ static void options(int *argc, char ***argv) {
 			die("Invalid --log-format option");
 	}
 
+	if (debug_srtp)
+		rtpe_config.common.log_levels[log_level_index_srtp] = LOG_DEBUG;
+
 	if (dtmf_udp_ep) {
 		if (endpoint_parse_any_getaddrinfo_full(&rtpe_config.dtmf_udp_ep, dtmf_udp_ep))
 			die("Invalid IP or port '%s' (--dtmf-log-dest)", dtmf_udp_ep);
@@ -603,8 +754,35 @@ static void options(int *argc, char ***argv) {
 	if (rtpe_config.dtls_rsa_key_size < 0)
 		die("Invalid --dtls-rsa-key-size (%i)", rtpe_config.dtls_rsa_key_size);
 
+	if (rtpe_config.dtls_mtu < 576)
+		/* The Internet Protocol requires that hosts must be able to process IP datagrams of at least 576 bytes (for IPv4) or 1280 bytes (for IPv6).
+		However, this does not preclude link layers with an MTU smaller than this minimum MTU from conveying IP data. Internet IPv4 path MTU is 68 bytes.*/
+		die("Invalid --dtls-mtu (%i)", rtpe_config.dtls_mtu);
+
 	if (rtpe_config.jb_length < 0)
 		die("Invalid negative jitter buffer size");
+
+	if (silence_detect > 0) {
+		rtpe_config.silence_detect_double = silence_detect / 100.0;
+		rtpe_config.silence_detect_int = (int) ((silence_detect / 100.0) * UINT32_MAX);
+	}
+
+	parse_cn_payload(&rtpe_config.cn_payload, cn_payload, "\x20", "cn-payload");
+	parse_cn_payload(&rtpe_config.dtx_cn_params, dtx_cn_params, NULL, "dtx-cn-params");
+
+	if (amr_dtx) {
+		if (!strcasecmp(amr_dtx, "native")) {}
+		else if (!strcasecmp(amr_dtx, "CN"))
+			rtpe_config.amr_cn_dtx = 1;
+		else
+			die("Invalid --amr-dtx ('%s')", amr_dtx);
+	}
+
+	if (!rtpe_config.software_id)
+		rtpe_config.software_id = g_strdup_printf("rtpengine-%s", RTPENGINE_VERSION);
+	g_strcanon(rtpe_config.software_id, "QWERTYUIOPASDFGHJKLZXCVBNMqwertyuiopasdfghjklzxcvbnm1234567890-", '-');
+
+	rwlock_unlock_w(&rtpe_config.config_lock);
 }
 
 void fill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
@@ -612,6 +790,8 @@ void fill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
 	GList* l;
 	struct intf_config* gptr_data;
 
+	rwlock_lock_w(&rtpe_config.config_lock);
+
 	for(l = rtpe_config.interfaces.head; l ; l=l->next) {
 		gptr_data = g_slice_alloc0(sizeof(*gptr_data));
 		memcpy(gptr_data, (struct intf_config*)(l->data), sizeof(*gptr_data));
@@ -657,12 +837,15 @@ void fill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
 	ini_rtpe_cfg->redis_disable_time = rtpe_config.redis_disable_time;
 	ini_rtpe_cfg->redis_cmd_timeout = rtpe_config.redis_cmd_timeout;
 	ini_rtpe_cfg->redis_connect_timeout = rtpe_config.redis_connect_timeout;
-	ini_rtpe_cfg->common.log_level = rtpe_config.common.log_level;
+	ini_rtpe_cfg->redis_delete_async = rtpe_config.redis_delete_async;
+	ini_rtpe_cfg->redis_delete_async_interval = rtpe_config.redis_delete_async_interval;
+	memcpy(&ini_rtpe_cfg->common.log_levels, &rtpe_config.common.log_levels, sizeof(ini_rtpe_cfg->common.log_levels));
 
 	ini_rtpe_cfg->graphite_ep = rtpe_config.graphite_ep;
 	ini_rtpe_cfg->tcp_listen_ep = rtpe_config.tcp_listen_ep;
 	ini_rtpe_cfg->udp_listen_ep = rtpe_config.udp_listen_ep;
 	ini_rtpe_cfg->ng_listen_ep = rtpe_config.ng_listen_ep;
+	ini_rtpe_cfg->ng_tcp_listen_ep = rtpe_config.ng_tcp_listen_ep;
 	ini_rtpe_cfg->cli_listen_ep = rtpe_config.cli_listen_ep;
 	ini_rtpe_cfg->redis_ep = rtpe_config.redis_ep;
 	ini_rtpe_cfg->redis_write_ep = rtpe_config.redis_write_ep;
@@ -679,6 +862,8 @@ void fill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
 
 	ini_rtpe_cfg->jb_length = rtpe_config.jb_length;
 	ini_rtpe_cfg->jb_clock_drift = rtpe_config.jb_clock_drift;
+
+	rwlock_unlock_w(&rtpe_config.config_lock);
 }
 
 static void
@@ -723,6 +908,15 @@ static void options_free(void) {
 	g_free(rtpe_config.mysql_pass);
 	g_free(rtpe_config.mysql_query);
 	g_free(rtpe_config.dtls_ciphers);
+	g_strfreev(rtpe_config.http_ifs);
+	g_strfreev(rtpe_config.https_ifs);
+	g_free(rtpe_config.https_cert);
+	g_free(rtpe_config.https_key);
+	g_free(rtpe_config.software_id);
+	if (rtpe_config.cn_payload.s)
+		free(rtpe_config.cn_payload.s);
+	if (rtpe_config.dtx_cn_params.s)
+		free(rtpe_config.dtx_cn_params.s);
 
 	// free common config options
 	config_load_free(&rtpe_config.common);
@@ -763,6 +957,7 @@ static void init_everything(void) {
 	dtmf_init();
 	jitter_buffer_init();
 	t38_init();
+	codecs_init();
 }
 
 
@@ -786,6 +981,10 @@ no_kernel:
 	if (!rtpe_poller)
 		die("poller creation failed");
 
+	rtpe_poller_map = poller_map_new();
+	if (!rtpe_poller_map)
+		die("poller map creation failed");
+
 	dtls_timer(rtpe_poller);
 
 	if (call_init())
@@ -796,14 +995,8 @@ no_kernel:
 		rtpe_config.max_sessions = -1;
 	}
 
-	if (rtpe_config.redis_num_threads < 1) {
-#ifdef _SC_NPROCESSORS_ONLN
-		rtpe_config.redis_num_threads = sysconf( _SC_NPROCESSORS_ONLN );
-#endif
-		if (rtpe_config.redis_num_threads < 1) {
-			rtpe_config.redis_num_threads = REDIS_RESTORE_NUM_THREADS;
-		}
-	}
+	if (rtpe_config.redis_num_threads < 1)
+		rtpe_config.redis_num_threads = num_cpu_cores(REDIS_RESTORE_NUM_THREADS);
 
 	rtpe_tcp = NULL;
 	if (rtpe_config.tcp_listen_ep.port) {
@@ -828,6 +1021,12 @@ no_kernel:
 			die("Failed to open UDP control connection port");
 	}
 
+	if (rtpe_config.ng_tcp_listen_ep.port) {
+		rtpe_control_ng = control_ng_tcp_new(rtpe_poller, &rtpe_config.ng_tcp_listen_ep, rtpe_control_ng);
+		if (!rtpe_control_ng)
+			die("Failed to open TCP control connection port");
+	}
+
 	rtpe_cli = NULL;
 	if (rtpe_config.cli_listen_ep.port) {
 		interfaces_exclude_port(rtpe_config.cli_listen_ep.port);
@@ -864,6 +1063,12 @@ no_kernel:
 			rtpe_redis_write = rtpe_redis;
 	}
 
+	if (rtpe_config.num_threads < 1)
+		rtpe_config.num_threads = num_cpu_cores(4);
+
+	if (websocket_init())
+		die("Failed to init websocket listener");
+
 	daemonize();
 	wpidfile();
 
@@ -876,8 +1081,22 @@ no_kernel:
 		gettimeofday(&redis_start, NULL);
 
 		// restore
-		if (redis_restore(rtpe_redis))
-			die("Refusing to continue without working Redis database");
+		if (rtpe_redis_notify) {
+			// active-active mode: the main DB has our own calls, while
+			// the "notify" DB has the "foreign" calls. "foreign" DB goes
+			// first as the "owned" DB can do a stray update back to Redis
+			for (GList *l = rtpe_config.redis_subscribed_keyspaces.head; l; l = l->next) {
+				int db = GPOINTER_TO_INT(l->data);
+				if (redis_restore(rtpe_redis_notify, 1, db))
+					ilog(LOG_WARN, "Unable to restore calls from the active-active peer");
+			}
+			if (redis_restore(rtpe_redis_write, 0, -1))
+				die("Refusing to continue without working Redis database");
+		}
+		else {
+			if (redis_restore(rtpe_redis, 0, -1))
+				die("Refusing to continue without working Redis database");
+		}
 
 		// stop redis restore timer
 		gettimeofday(&redis_stop, NULL);
@@ -905,41 +1124,50 @@ int main(int argc, char **argv) {
 
 	ilog(LOG_INFO, "Startup complete, version %s", RTPENGINE_VERSION);
 
-	thread_create_detach(sighandler, NULL);
-	thread_create_detach_prio(poller_timer_loop, rtpe_poller, rtpe_config.idle_scheduling, rtpe_config.idle_priority);
-	thread_create_detach_prio(load_thread, NULL, rtpe_config.idle_scheduling, rtpe_config.idle_priority);
+	thread_create_detach(sighandler, NULL, "signal handler");
+	thread_create_detach_prio(poller_timer_loop, rtpe_poller, rtpe_config.idle_scheduling,
+			rtpe_config.idle_priority, "poller timer");
+	thread_create_detach_prio(load_thread, NULL, rtpe_config.idle_scheduling, rtpe_config.idle_priority, "load monitor");
+
+	if (!is_addr_unspecified(&rtpe_config.redis_ep.address) && initial_rtpe_config.redis_delete_async)
+		thread_create_detach(redis_delete_async_loop, NULL, "redis async");
 
 	if (!is_addr_unspecified(&rtpe_config.redis_ep.address) && rtpe_redis_notify)
-		thread_create_detach(redis_notify_loop, NULL);
+		thread_create_detach(redis_notify_loop, NULL, "redis notify");
 
 	if (!is_addr_unspecified(&rtpe_config.graphite_ep.address))
-		thread_create_detach(graphite_loop, NULL);
+		thread_create_detach(graphite_loop, NULL, "graphite");
 
-	thread_create_detach(ice_thread_run, NULL);
+	thread_create_detach(ice_thread_run, NULL, "ICE");
 
-	if (rtpe_config.num_threads < 1) {
-#ifdef _SC_NPROCESSORS_ONLN
-		rtpe_config.num_threads = sysconf( _SC_NPROCESSORS_ONLN ) + 3;
-#endif
-		if (rtpe_config.num_threads <= 1)
-			rtpe_config.num_threads = 4;
-	}
+	websocket_start();
 
 	service_notify("READY=1\n");
 
-	for (idx = 0; idx < rtpe_config.num_threads; ++idx)
-		thread_create_detach_prio(poller_loop, rtpe_poller, rtpe_config.scheduling, rtpe_config.priority);
+	for (idx = 0; idx < rtpe_config.num_threads; ++idx) {
+		if (!rtpe_config.poller_per_thread)
+			thread_create_detach_prio(poller_loop2, rtpe_poller, rtpe_config.scheduling, rtpe_config.priority, "poller");
+		else
+			thread_create_detach_prio(poller_loop, rtpe_poller_map, rtpe_config.scheduling, rtpe_config.priority, "poller");
+	}
+
+	if (!rtpe_config.poller_per_thread)
+		thread_create_detach_prio(poller_loop2, rtpe_poller, rtpe_config.scheduling, rtpe_config.priority, "poller");
 
 	if (rtpe_config.media_num_threads < 0)
 		rtpe_config.media_num_threads = rtpe_config.num_threads;
 	for (idx = 0; idx < rtpe_config.media_num_threads; ++idx) {
 #ifdef WITH_TRANSCODING
-		thread_create_detach_prio(media_player_loop, NULL, rtpe_config.scheduling, rtpe_config.priority);
+		thread_create_detach_prio(media_player_loop, NULL, rtpe_config.scheduling,
+				rtpe_config.priority, "media player");
 #endif
-		thread_create_detach_prio(send_timer_loop, NULL, rtpe_config.scheduling, rtpe_config.priority);
+		thread_create_detach_prio(send_timer_loop, NULL, rtpe_config.scheduling,
+				rtpe_config.priority, "send timer");
 		if (rtpe_config.jb_length > 0)
 			thread_create_detach_prio(jitter_buffer_loop, NULL, rtpe_config.scheduling,
-					rtpe_config.priority);
+					rtpe_config.priority, "jitter buffer");
+		thread_create_detach_prio(codec_timers_loop, NULL, rtpe_config.scheduling,
+				rtpe_config.priority, "codec timer");
 	}
 
 
@@ -948,15 +1176,26 @@ int main(int argc, char **argv) {
 		threads_join_all(0);
 	}
 
+        // free libevent
+#if LIBEVENT_VERSION_NUMBER >= 0x02010100
+        libevent_global_shutdown();
+#endif
+
 	service_notify("STOPPING=1\n");
 
+	if (!is_addr_unspecified(&rtpe_config.redis_ep.address) && initial_rtpe_config.redis_delete_async)
+		redis_async_event_base_action(rtpe_redis_write, EVENT_BASE_LOOPBREAK);
+
 	if (!is_addr_unspecified(&rtpe_config.redis_ep.address) && rtpe_redis_notify)
-		redis_notify_event_base_action(EVENT_BASE_LOOPBREAK);
+		redis_async_event_base_action(rtpe_redis_notify, EVENT_BASE_LOOPBREAK);
 
 	threads_join_all(1);
 
+	if (!is_addr_unspecified(&rtpe_config.redis_ep.address) && initial_rtpe_config.redis_delete_async)
+		redis_async_event_base_action(rtpe_redis_write, EVENT_BASE_FREE);
+
 	if (!is_addr_unspecified(&rtpe_config.redis_ep.address) && rtpe_redis_notify)
-		redis_notify_event_base_action(EVENT_BASE_FREE);
+		redis_async_event_base_action(rtpe_redis_notify, EVENT_BASE_FREE);
 
 	ilog(LOG_INFO, "Version %s shutting down", RTPENGINE_VERSION);
 
@@ -969,14 +1208,16 @@ int main(int argc, char **argv) {
 	jitter_buffer_init_free();
 	media_player_free();
 	codeclib_free();
-	statistics_free();
 	call_interfaces_free();
-	interfaces_free();
 	ice_free();
 	dtls_cert_free();
+	control_ng_cleanup();
+	codecs_cleanup();
+	statistics_free();
 
 	redis_close(rtpe_redis);
-	redis_close(rtpe_redis_write);
+	if (rtpe_redis_write != rtpe_redis)
+		redis_close(rtpe_redis_write);
 	redis_close(rtpe_redis_notify);
 
 	free_prefix();
@@ -990,6 +1231,8 @@ int main(int argc, char **argv) {
 	obj_release(rtpe_tcp);
 	obj_release(rtpe_control_ng);
 	poller_free(&rtpe_poller);
+	poller_map_free(&rtpe_poller_map);
+	interfaces_free();
 
 	return 0;
 }
