@@ -4,6 +4,7 @@
 #include <glib.h>
 #include <errno.h>
 #include <netinet/in.h>
+#include <stdbool.h>
 #include "str.h"
 #include "ice.h"
 #include "socket.h"
@@ -911,40 +912,84 @@ fail:
 }
 
 /* puts a list of "struct intf_list" into "out", containing socket_t list */
-int get_consecutive_ports(GQueue *out, unsigned int num_ports, const struct logical_intf *log,
-		const str *label)
+int get_consecutive_ports(GQueue *out, unsigned int num_ports, struct call_media *media)
 {
 	GList *l;
 	struct intf_list *il;
 	const struct local_intf *loc;
+	const struct logical_intf *log = media->logical_intf;
+	const sockfamily_t *desired_family = media->desired_family;
+	const str *label = &media->call->callid;
 
+	/*
+	// debug locals of logical incerface
+	char ip[100];
 	for (l = log->list.head; l; l = l->next) {
 		loc = l->data;
+		inet_ntop(loc->spec->local_address.addr.family->af, &loc->spec->local_address.addr.u, ip, sizeof(ip));
+		ilog(LOG_DEBUG, "XXXXXXXXXX IP: %s", ip);
+	}
+	ilog(LOG_DEBUG, "");
+	*/
 
-		il = g_slice_alloc0(sizeof(*il));
-		il->local_intf = loc;
-		g_queue_push_tail(out, il);
-		if (G_LIKELY(!__get_consecutive_ports(&il->list, num_ports, 0, loc->spec, label))) {
-			// success - found available ports on local interfaces, so far
-			continue;
+	if (!rtpe_config.save_interface_ports) {
+		for (l = log->list.head; l; l = l->next) {
+			loc = l->data;
+
+			il = g_slice_alloc0(sizeof(*il));
+			il->local_intf = loc;
+			g_queue_push_tail(out, il);
+			if (G_LIKELY(!__get_consecutive_ports(&il->list, num_ports, 0, loc->spec, label))) {
+				// success - found available ports on local interfaces, so far
+				continue;
+			} else {
+				// fail - did not found available ports on at least one local interface
+				goto error_ports;
+			}
 		}
 
-		// error - found at least one local interface with no ports available
-		goto error_ports;
-	}
-
-	return 0;
+		return 0;
 
 error_ports:
-	ilog(LOG_ERR, "Failed to get %d consecutive ports on all locals of logical '"STR_FORMAT"'",
-		num_ports, STR_FMT(&log->name));
+		ilog(LOG_ERR, "Failed to get %d consecutive ports on all locals of logical '"STR_FORMAT"'",
+			num_ports, STR_FMT(&log->name));
 
-	// free all ports alloc'ed so far for the previous local interfaces
-	while ((il = g_queue_pop_head(out))) {
-		free_socket_intf_list(il);
+		// free all ports alloc'ed so far for the previous local interfaces
+		while ((il = g_queue_pop_head(out))) {
+			free_socket_intf_list(il);
+		}
+
+		return -1;
+
+	} else {
+		for (l = log->list.head; l; l = l->next) {
+			loc = l->data;
+
+			// check desired family of local interface
+			if (desired_family != loc->spec->local_address.addr.family)  {
+				ilog(LOG_DEBUG, "Did not find yet one local interface for family %s; continue...", desired_family->rfc_name);
+				continue;
+			}
+
+			ilog(LOG_DEBUG, "Found one local interface for family %s", desired_family->rfc_name);
+
+			il = g_slice_alloc0(sizeof(*il));
+			il->local_intf = loc;
+			if (G_LIKELY(!__get_consecutive_ports(&il->list, num_ports, 0, loc->spec, label))) {
+				// success - found available ports on one local interface
+				g_queue_push_tail(out, il);
+				return 0;
+			} else {
+				// fail - no available ports on one local interface... continue
+				free_socket_intf_list(il);
+			}
+		}
+
+		ilog(LOG_ERR, "Failed to get %d consecutive ports on one local of logical '"STR_FORMAT"'",
+			num_ports, STR_FMT(&log->name));
+
+		return -1;
 	}
-
-	return -1;
 }
 void free_socket_intf_list(struct intf_list *il) {
 	socket_t *sock;
@@ -1033,6 +1078,8 @@ static int __k_srtp_crypt(struct rtpengine_srtp *s, struct crypto_context *c, st
 	s->master_key_len = c->params.crypto_suite->master_key_len;
 	s->session_key_len = c->params.crypto_suite->session_key_len;
 	memcpy(s->master_salt, c->params.master_salt, c->params.crypto_suite->master_salt_len);
+	s->master_salt_len = c->params.crypto_suite->master_salt_len;
+	s->session_salt_len = c->params.crypto_suite->session_salt_len;
 
 	if (c->params.session_params.unencrypted_srtp)
 		s->cipher = REC_NULL;
@@ -1095,6 +1142,8 @@ void kernelize(struct packet_stream *stream) {
 		goto no_kernel;
 	if (!stream->endpoint.address.family)
 		goto no_kernel;
+	if (MEDIA_ISSET(media, BLACKHOLE))
+		non_forwarding = 1;
 
         ilog(LOG_INFO, "Kernelizing media stream: %s%s:%d%s",
 			FMT_M(sockaddr_print_buf(&stream->endpoint.address), stream->endpoint.port));
@@ -1137,12 +1186,14 @@ void kernelize(struct packet_stream *stream) {
 	reti.dtls = MEDIA_ISSET(media, DTLS);
 	reti.stun = media->ice_agent ? 1 : 0;
 	reti.non_forwarding = non_forwarding;
+	reti.blackhole = MEDIA_ISSET(media, BLACKHOLE) ? 1 : 0;
+	reti.rtp_stats = MEDIA_ISSET(media, RTCP_GEN) ? 1 : 0;
 
 	__re_address_translate_ep(&reti.dst_addr, &sink->endpoint);
 	__re_address_translate_ep(&reti.src_addr, &sink->selected_sfd->socket.local);
 	if (stream->ssrc_in) {
 		reti.ssrc = htonl(stream->ssrc_in->parent->h.ssrc);
-		if (MEDIA_ISSET(media, TRANSCODE)) {
+		if (MEDIA_ISSET(media, TRANSCODE) || MEDIA_ISSET(media, ECHO)) {
 			reti.ssrc_out = htonl(stream->ssrc_in->ssrc_map_out);
 			reti.transcoding = 1;
 		}
@@ -1179,7 +1230,9 @@ void kernelize(struct packet_stream *stream) {
 			struct codec_handler *ch = codec_handler_get(media, rs->payload_type);
 			if (!ch->kernelize)
 				continue;
-			reti.payload_types[reti.num_payload_types++] = rs->payload_type;
+			reti.payload_types[reti.num_payload_types] = rs->payload_type;
+			reti.clock_rates[reti.num_payload_types] = ch->source_pt.clock_rate;
+			reti.num_payload_types++;
 		}
 		g_list_free(values);
 	}
@@ -1190,7 +1243,7 @@ void kernelize(struct packet_stream *stream) {
 
 	recording_stream_kernel_info(stream, &reti);
 
-	kernel_add_stream(&reti, 0);
+	kernel_add_stream(&reti);
 	PS_SET(stream, KERNELIZED);
 
 	return;
@@ -1202,6 +1255,67 @@ no_kernel:
 	PS_SET(stream, NO_KERNEL_SUPPORT);
 }
 
+// must be called with appropriate locks (master lock and/or in_lock)
+static void __stream_update_stats(struct packet_stream *ps, int have_in_lock) {
+	struct re_address local;
+
+	if (!have_in_lock)
+		mutex_lock(&ps->in_lock);
+
+	struct ssrc_ctx *ssrc_ctx = ps->ssrc_in;
+	if (!ssrc_ctx) {
+		if (!have_in_lock)
+			mutex_unlock(&ps->in_lock);
+		return;
+	}
+	struct ssrc_entry_call *parent = ssrc_ctx->parent;
+
+	__re_address_translate_ep(&local, &ps->selected_sfd->socket.local);
+	struct rtpengine_ssrc_stats stats;
+	if (kernel_update_stats(&local, htonl(parent->h.ssrc), &stats)) {
+		if (!have_in_lock)
+			mutex_unlock(&ps->in_lock);
+		return;
+	}
+
+	if (!stats.basic_stats.packets) {
+		// no change
+		if (!have_in_lock)
+			mutex_unlock(&ps->in_lock);
+		return;
+	}
+
+	atomic64_add(&ssrc_ctx->packets, stats.basic_stats.packets);
+	atomic64_add(&ssrc_ctx->octets, stats.basic_stats.bytes);
+	atomic64_add(&ssrc_ctx->packets_lost, stats.total_lost);
+	atomic64_set(&ssrc_ctx->last_seq, stats.ext_seq);
+	atomic64_set(&ssrc_ctx->last_ts, stats.timestamp);
+	parent->jitter = stats.jitter;
+
+	uint32_t ssrc_map_out = ssrc_ctx->ssrc_map_out;
+
+	if (!have_in_lock)
+		mutex_unlock(&ps->in_lock);
+
+	// update opposite outgoing SSRC
+	if (!have_in_lock)
+		mutex_lock(&ps->out_lock);
+	else {
+		if (mutex_trylock(&ps->out_lock))
+			return; // will have to skip this
+	}
+	ssrc_ctx = ps->ssrc_out;
+	if (ssrc_ctx) {
+		parent = ssrc_ctx->parent;
+		if (parent->h.ssrc == ssrc_map_out) {
+			atomic64_add(&ssrc_ctx->packets, stats.basic_stats.packets);
+			atomic64_add(&ssrc_ctx->octets, stats.basic_stats.bytes);
+		}
+	}
+	mutex_unlock(&ps->out_lock);
+}
+
+
 /* must be called with in_lock held or call->master_lock held in W */
 void __unkernelize(struct packet_stream *p) {
 	struct re_address rea;
@@ -1212,6 +1326,7 @@ void __unkernelize(struct packet_stream *p) {
 		return;
 
 	if (kernel.is_open) {
+		__stream_update_stats(p, 1);
 		__re_address_translate_ep(&rea, &p->selected_sfd->socket.local);
 		kernel_del_stream(&rea);
 	}
@@ -1241,6 +1356,24 @@ void unkernelize(struct packet_stream *ps) {
 	mutex_unlock(&ps->in_lock);
 }
 
+// master lock held in R
+void media_update_stats(struct call_media *m) {
+	if (!proto_is_rtp(m->protocol))
+		return;
+	if (!kernel.is_open)
+		return;
+
+	for (GList *l = m->streams.head; l; l = l->next) {
+		struct packet_stream *ps = l->data;
+		if (!PS_ISSET(ps, RTP))
+			continue;
+		if (!PS_ISSET(ps, KERNELIZED))
+			continue;
+
+		__stream_update_stats(ps, 0);
+	}
+}
+
 
 
 const struct streamhandler *determine_handler(const struct transport_protocol *in_proto,
@@ -1319,11 +1452,19 @@ noop:
 
 
 // check and update SSRC pointers
-static void __stream_ssrc(struct packet_stream *in_srtp, struct packet_stream *out_srtp, u_int32_t ssrc_bs,
-		struct ssrc_ctx **ssrc_in_p, struct ssrc_ctx **ssrc_out_p, struct ssrc_hash *ssrc_hash, struct packet_handler_ctx *phc)
+static void __stream_ssrc(struct packet_stream *in_srtp, struct packet_stream *out_srtp, uint32_t ssrc_bs,
+		struct packet_handler_ctx *phc)
 {
-	u_int32_t in_ssrc = ntohl(ssrc_bs);
-	u_int32_t out_ssrc;
+	uint32_t in_ssrc = ntohl(ssrc_bs);
+	uint32_t out_ssrc;
+	struct ssrc_ctx **ssrc_in_p = &phc->mp.ssrc_in;
+	struct ssrc_ctx **ssrc_out_p = &phc->mp.ssrc_out;
+
+	if (!phc->mp.media || !phc->mp.media_out)
+		return;
+
+	struct ssrc_hash *ssrc_hash_in = phc->mp.media->monologue->ssrc_hash;
+	struct ssrc_hash *ssrc_hash_out = phc->mp.media_out->monologue->ssrc_hash;
 
 	// input direction
 	mutex_lock(&in_srtp->in_lock);
@@ -1335,7 +1476,7 @@ static void __stream_ssrc(struct packet_stream *in_srtp, struct packet_stream *o
 		ssrc_ctx_put(ssrc_in_p);
 		ssrc_ctx_put(&in_srtp->ssrc_in);
 		(*ssrc_in_p) = in_srtp->ssrc_in =
-			get_ssrc_ctx(in_ssrc, ssrc_hash, SSRC_DIR_INPUT, in_srtp->media->monologue);
+			get_ssrc_ctx(in_ssrc, ssrc_hash_in, SSRC_DIR_INPUT, in_srtp->media->monologue);
 		ssrc_ctx_hold(in_srtp->ssrc_in);
 
 		phc->unkernelize = 1;
@@ -1344,7 +1485,7 @@ static void __stream_ssrc(struct packet_stream *in_srtp, struct packet_stream *o
 	}
 
 	// make sure we reset the output SSRC if we're not transcoding
-	if (!MEDIA_ISSET(in_srtp->media, TRANSCODE))
+	if (!MEDIA_ISSET(in_srtp->media, TRANSCODE) && !MEDIA_ISSET(in_srtp->media, ECHO))
 		(*ssrc_in_p)->ssrc_map_out = in_ssrc;
 
 	out_ssrc = (*ssrc_in_p)->ssrc_map_out;
@@ -1360,9 +1501,10 @@ static void __stream_ssrc(struct packet_stream *in_srtp, struct packet_stream *o
 		ssrc_ctx_put(ssrc_out_p);
 		ssrc_ctx_put(&out_srtp->ssrc_out);
 		(*ssrc_out_p) = out_srtp->ssrc_out =
-			get_ssrc_ctx(out_ssrc, ssrc_hash, SSRC_DIR_OUTPUT, out_srtp->media->monologue);
+			get_ssrc_ctx(out_ssrc, ssrc_hash_out, SSRC_DIR_OUTPUT, out_srtp->media->monologue);
 		ssrc_ctx_hold(out_srtp->ssrc_out);
 
+		// coverity[missing_lock : FALSE]
 		phc->unkernelize = 1;
 		ilog(LOG_DEBUG, ">>> out_ssrc changed for %s%s:%d new: %x %s",
                         FMT_M(sockaddr_print_buf(&out_srtp->endpoint.address), out_srtp->endpoint.port, out_ssrc));
@@ -1395,8 +1537,9 @@ static int media_demux_protocols(struct packet_handler_ctx *phc) {
 			call_media_state_machine(phc->mp.media);
 			return 1;
 		}
-		else /* not an stun packet */
-			;
+		else {
+			/* not an stun packet */
+		}
 	}
 	return -1;
 }
@@ -1416,8 +1559,10 @@ static int media_loop_detect(struct packet_handler_ctx *phc) {
 
 		__C_DBG("packet dupe");
 		if (phc->mp.stream->lp_count >= RTP_LOOP_MAX_COUNT) {
-			ilog(LOG_WARNING, "More than %d duplicate packets detected, dropping packet "
-					"to avoid potential loop", RTP_LOOP_MAX_COUNT);
+			ilog(LOG_WARNING, "More than %d duplicate packets detected, dropping packet from %s%s%s"
+					"to avoid potential loop",
+					RTP_LOOP_MAX_COUNT,
+					FMT_M(endpoint_print_buf(&phc->mp.fsin)));
 			mutex_unlock(&phc->mp.stream->in_lock);
 			return -1;
 		}
@@ -1459,8 +1604,12 @@ static void media_packet_rtcp_demux(struct packet_handler_ctx *phc)
 		}
 	}
 	phc->out_srtp = phc->sink;
-	if (phc->rtcp && phc->sink && phc->sink->rtcp_sibling)
+	if (!phc->sink)
+		return;
+	if (phc->rtcp && phc->sink->rtcp_sibling)
 		phc->out_srtp = phc->sink->rtcp_sibling; // use RTCP SRTP context
+
+	phc->mp.media_out = phc->sink->media;
 }
 
 
@@ -1468,6 +1617,8 @@ static void media_packet_rtp(struct packet_handler_ctx *phc)
 {
 	phc->payload_type = -1;
 
+	if (G_UNLIKELY(!phc->mp.media))
+		return;
 	if (G_UNLIKELY(!proto_is_rtp(phc->mp.media->protocol)))
 		return;
 
@@ -1475,8 +1626,7 @@ static void media_packet_rtp(struct packet_handler_ctx *phc)
 		rtp_padding(phc->mp.rtp, &phc->mp.payload);
 
 		if (G_LIKELY(phc->out_srtp != NULL))
-			__stream_ssrc(phc->in_srtp, phc->out_srtp, phc->mp.rtp->ssrc, &phc->mp.ssrc_in,
-					&phc->mp.ssrc_out, phc->mp.call->ssrc_hash, phc);
+			__stream_ssrc(phc->in_srtp, phc->out_srtp, phc->mp.rtp->ssrc, phc);
 
 		// check the payload type
 		// XXX redundant between SSRC handling and codec_handler stuff -> combine
@@ -1487,10 +1637,12 @@ static void media_packet_rtp(struct packet_handler_ctx *phc)
 		// XXX yet another hash table per payload type -> combine
 		struct rtp_stats *rtp_s = g_atomic_pointer_get(&phc->mp.stream->rtp_stats_cache);
 		if (G_UNLIKELY(!rtp_s) || G_UNLIKELY(rtp_s->payload_type != phc->payload_type))
-			rtp_s = g_hash_table_lookup(phc->mp.stream->rtp_stats, &phc->payload_type);
+			rtp_s = g_hash_table_lookup(phc->mp.stream->rtp_stats, GINT_TO_POINTER(phc->payload_type));
 		if (!rtp_s) {
 			ilog(LOG_WARNING | LOG_FLAG_LIMIT,
-					"RTP packet with unknown payload type %u received", phc->payload_type);
+					"RTP packet with unknown payload type %u received from %s%s%s",
+					phc->payload_type,
+					FMT_M(endpoint_print_buf(&phc->mp.fsin)));
 			atomic64_inc(&phc->mp.stream->stats.errors);
 			atomic64_inc(&rtpe_statsps.errors);
 		}
@@ -1502,8 +1654,7 @@ static void media_packet_rtp(struct packet_handler_ctx *phc)
 	}
 	else if (phc->rtcp && !rtcp_payload(&phc->mp.rtcp, NULL, &phc->s)) {
 		if (G_LIKELY(phc->out_srtp != NULL))
-			__stream_ssrc(phc->in_srtp, phc->out_srtp, phc->mp.rtcp->ssrc, &phc->mp.ssrc_in,
-					&phc->mp.ssrc_out, phc->mp.call->ssrc_hash, phc);
+			__stream_ssrc(phc->in_srtp, phc->out_srtp, phc->mp.rtcp->ssrc, phc);
 	}
 }
 
@@ -1598,8 +1749,9 @@ static int media_packet_address_check(struct packet_handler_ctx *phc)
 	// is enabled or not.
 	if (!PS_ISSET(phc->mp.stream, CONFIRMED) && PS_ISSET(phc->mp.stream, RTP)) {
 		if (rtcp_demux_is_rtcp(&phc->s)) {
-			ilog(LOG_DEBUG | LOG_FLAG_LIMIT, "Ignoring stray RTCP packet for "
-					"peer address confirmation purposes");
+			ilog(LOG_DEBUG | LOG_FLAG_LIMIT, "Ignoring stray RTCP packet from %s%s%s for "
+					"peer address confirmation purposes",
+					FMT_M(endpoint_print_buf(&phc->mp.fsin)));
 			goto out;
 		}
 	}
@@ -1647,6 +1799,12 @@ static int media_packet_address_check(struct packet_handler_ctx *phc)
 		goto out;
 	}
 
+	/* wait at least 3 seconds after last signal before committing to a particular
+	 * endpoint address */
+	bool wait_time = false;
+	if (!phc->mp.call->last_signal || rtpe_now.tv_sec <= phc->mp.call->last_signal + 3)
+		wait_time = true;
+
 	const struct endpoint *use_endpoint_confirm = &phc->mp.fsin;
 
 	if (rtpe_config.endpoint_learning == EL_IMMEDIATE)
@@ -1685,9 +1843,7 @@ static int media_packet_address_check(struct packet_handler_ctx *phc)
 		}
 	}
 
-	/* wait at least 3 seconds after last signal before committing to a particular
-	 * endpoint address */
-	if (!phc->mp.call->last_signal || rtpe_now.tv_sec <= phc->mp.call->last_signal + 3)
+	if (wait_time)
 		goto update_peerinfo;
 
 confirm_now:
@@ -1700,11 +1856,18 @@ confirm_now:
 
 update_peerinfo:
 	mutex_lock(&phc->mp.stream->out_lock);
-	endpoint = phc->mp.stream->endpoint;
-	phc->mp.stream->endpoint = *use_endpoint_confirm;
-	if (memcmp(&endpoint, &phc->mp.stream->endpoint, sizeof(endpoint))) {
-		phc->unkernelize = 1;
-		phc->update = 1;
+	// if we're during the wait time, check the received address against the previously
+	// learned address. if they're the same, ignore this packet for learning purposes
+	if (!wait_time || !phc->mp.stream->learned_endpoint.address.family ||
+			memcmp(use_endpoint_confirm, &phc->mp.stream->learned_endpoint, sizeof(endpoint)))
+	{
+		endpoint = phc->mp.stream->endpoint;
+		phc->mp.stream->endpoint = *use_endpoint_confirm;
+		phc->mp.stream->learned_endpoint = *use_endpoint_confirm;
+		if (memcmp(&endpoint, &phc->mp.stream->endpoint, sizeof(endpoint))) {
+			phc->unkernelize = 1;
+			phc->update = 1;
+		}
 	}
 update_addr:
 	mutex_unlock(&phc->mp.stream->out_lock);
@@ -1769,16 +1932,19 @@ static int do_rtcp(struct packet_handler_ctx *phc) {
 	int ret = -1;
 
 	GQueue rtcp_list = G_QUEUE_INIT;
-	if (rtcp_parse(&rtcp_list, &phc->mp))
+	int rtcp_ret = rtcp_parse(&rtcp_list, &phc->mp);
+	if (rtcp_ret < 0)
 		goto out;
+	if (rtcp_ret == 1)
+		goto ok;
 	if (phc->rtcp_filter)
 		if (phc->rtcp_filter(&phc->mp, &rtcp_list))
 			goto out;
 
 	// queue for output
 	codec_add_raw_packet(&phc->mp);
+ok:
 	ret = 0;
-
 out:
 	rtcp_list_free(&rtcp_list);
 	return ret;
@@ -1789,7 +1955,7 @@ out:
 int media_socket_dequeue(struct media_packet *mp, struct packet_stream *sink) {
 	struct codec_packet *p;
 	while ((p = g_queue_pop_head(&mp->packets_out))) {
-		if (sink->send_timer)
+		if (sink && sink->send_timer)
 			send_timer_push(sink->send_timer, p);
 		else
 			codec_packet_free(p);
@@ -1797,6 +1963,32 @@ int media_socket_dequeue(struct media_packet *mp, struct packet_stream *sink) {
 	return 0;
 }
 
+void media_packet_copy(struct media_packet *dst, const struct media_packet *src) {
+	*dst = *src;
+	g_queue_init(&dst->packets_out);
+	if (dst->sfd)
+		obj_hold(dst->sfd);
+	if (dst->ssrc_in)
+		obj_hold(&dst->ssrc_in->parent->h);
+	if (dst->ssrc_out)
+		obj_hold(&dst->ssrc_out->parent->h);
+	dst->rtp = __g_memdup(src->rtp, sizeof(*src->rtp));
+	dst->rtcp = __g_memdup(src->rtp, sizeof(*src->rtp));
+	dst->payload = STR_NULL;
+	dst->raw = STR_NULL;
+}
+void media_packet_release(struct media_packet *mp) {
+	if (mp->sfd)
+		obj_put(mp->sfd);
+	if (mp->ssrc_in)
+		obj_put(&mp->ssrc_in->parent->h);
+	if (mp->ssrc_out)
+		obj_put(&mp->ssrc_out->parent->h);
+	g_queue_clear_full(&mp->packets_out, codec_packet_free);
+	g_free(mp->rtp);
+	g_free(mp->rtcp);
+}
+
 
 /* called lock-free */
 static int stream_packet(struct packet_handler_ctx *phc) {
@@ -1838,7 +2030,9 @@ static int stream_packet(struct packet_handler_ctx *phc) {
 	if (!phc->mp.stream->selected_sfd)
 		goto out;
 
-	if (phc->mp.call && phc->mp.call->drop_traffic) {
+	phc->mp.call->foreign_media = 0;
+
+	if (phc->mp.call->drop_traffic) {
 		goto drop;
 	}
 
@@ -1858,7 +2052,7 @@ static int stream_packet(struct packet_handler_ctx *phc) {
 	}
 #endif
 
-	if (IS_FOREIGN_CALL(phc->mp.call))
+	if (rtpe_config.active_switchover && IS_FOREIGN_CALL(phc->mp.call))
 		call_make_own_foreign(phc->mp.call, 0);
 
 	// this sets rtcp, in_srtp, out_srtp, and sink
@@ -1892,7 +2086,8 @@ static int stream_packet(struct packet_handler_ctx *phc) {
 	if (G_UNLIKELY(!phc->sink || !phc->sink->selected_sfd || !phc->out_srtp
 				|| !phc->out_srtp->selected_sfd || !phc->in_srtp->selected_sfd))
 	{
-		ilog(LOG_WARNING, "Media packet from %s%s%s discarded", FMT_M(endpoint_print_buf(&phc->mp.fsin)));
+		ilog(LOG_WARNING | LOG_FLAG_LIMIT, "Media packet from %s%s%s discarded due to lack of sink",
+				FMT_M(endpoint_print_buf(&phc->mp.fsin)));
 		atomic64_inc(&phc->mp.stream->stats.errors);
 		atomic64_inc(&rtpe_statsps.errors);
 		goto out;
@@ -1944,7 +2139,10 @@ static int stream_packet(struct packet_handler_ctx *phc) {
 		goto drop;
 	}
 
-	ret = media_socket_dequeue(&phc->mp, phc->sink);
+	if (!MEDIA_ISSET(phc->mp.media, BLACKHOLE))
+		ret = media_socket_dequeue(&phc->mp, phc->sink);
+	else
+		ret = media_socket_dequeue(&phc->mp, NULL);
 
 	mutex_unlock(&phc->sink->out_lock);
 
@@ -2064,6 +2262,7 @@ static void stream_fd_free(void *p) {
 struct stream_fd *stream_fd_new(socket_t *fd, struct call *call, const struct local_intf *lif) {
 	struct stream_fd *sfd;
 	struct poller_item pi;
+	struct poller *p = rtpe_poller;
 
 	sfd = obj_alloc0("stream_fd", sizeof(*sfd), stream_fd_free);
 	sfd->unique_id = g_queue_get_length(&call->stream_fds);
@@ -2081,8 +2280,12 @@ struct stream_fd *stream_fd_new(socket_t *fd, struct call *call, const struct lo
 	pi.readable = stream_fd_readable;
 	pi.closed = stream_fd_closed;
 
-	if (poller_add_item(rtpe_poller, &pi))
-		ilog(LOG_ERR, "Failed to add stream_fd to poller");
+	if (rtpe_config.poller_per_thread)
+		p = poller_map_get(rtpe_poller_map);
+	if (p) {
+		if (poller_add_item(p, &pi))
+			ilog(LOG_ERR, "Failed to add stream_fd to poller");
+	}
 
 	return sfd;
 }
