@@ -37,10 +37,6 @@ struct redis		*rtpe_redis;
 struct redis		*rtpe_redis_write;
 struct redis		*rtpe_redis_notify;
 
-struct event_base	*rtpe_redis_notify_event_base;
-struct redisAsyncContext *rtpe_redis_notify_async_context;
-
-
 
 INLINE redisReply *redis_expect(int type, redisReply *r) {
 	if (!r)
@@ -82,6 +78,7 @@ static int redisCommandNR(redisContext *r, const char *fmt, ...)
 static int redis_check_conn(struct redis *r);
 static void json_restore_call(struct redis *r, const str *id, int foreign);
 static int redis_connect(struct redis *r, int wait);
+static int json_build_ssrc(struct call_monologue *ml, JsonReader *root_reader);
 
 static void redis_pipe(struct redis *r, const char *fmt, ...) {
 	va_list ap;
@@ -197,6 +194,16 @@ int redis_reconnect(struct redis* r) {
 	return rval;
 }
 
+// struct must be locked or single thread
+static int redis_select_db(struct redis *r, int db) {
+	if (db == r->current_db)
+		return 0;
+	if (redisCommandNR(r->ctx, "SELECT %i", db))
+		return -1;
+	r->current_db = db;
+	return 0;
+}
+
 /* called with r->lock held if necessary */
 static int redis_connect(struct redis *r, int wait) {
 	struct timeval tv;
@@ -207,6 +214,7 @@ static int redis_connect(struct redis *r, int wait) {
 	if (r->ctx)
 		redisFree(r->ctx);
 	r->ctx = NULL;
+	r->current_db = -1;
 
 	rwlock_lock_r(&rtpe_config.config_lock);
 	connect_timeout = rtpe_config.redis_connect_timeout;
@@ -234,7 +242,7 @@ static int redis_connect(struct redis *r, int wait) {
 			goto err2;
 	}
 
-	if (redisCommandNR(r->ctx, "SELECT %i", r->db))
+	if (redis_select_db(r, r->db))
 		goto err2;
 
 	while (wait-- >= 0) {
@@ -352,7 +360,7 @@ void on_redis_notification(redisAsyncContext *actx, void *reply, void *privdata)
 		goto err;
 
 	// select the right db for restoring the call
-	if (redisCommandNR(r->ctx, "SELECT %i", r->db)) {
+	if (redis_select_db(r, r->db)) {
 		if (r->ctx && r->ctx->err)
 			rlog(LOG_ERROR, "Redis error: %s", r->ctx->errstr);
 		redisFree(r->ctx);
@@ -371,7 +379,12 @@ void on_redis_notification(redisAsyncContext *actx, void *reply, void *privdata)
 				goto err;
 			}
 		}
+	        mutex_unlock(&r->lock);
+
+		// unlock before restoring calls to avoid deadlock in case err happens
 		json_restore_call(r, &callid, 1);
+
+	        mutex_lock(&r->lock);
 	}
 
 	if (strncmp(rr->element[3]->str,"del",3)==0) {
@@ -398,119 +411,168 @@ err:
 	mutex_unlock(&r->lock);
 }
 
-void redis_async_context_disconnect(const redisAsyncContext *redis_notify_async_context, int status) {
+void redis_delete_async_context_connect(const redisAsyncContext *redis_delete_async_context, int status) {
 	if (status == REDIS_ERR) {
-		if (redis_notify_async_context->errstr) {
-			rlog(LOG_ERROR, "redis_async_context_disconnect error %d on context free: %s",
-				redis_notify_async_context->err, redis_notify_async_context->errstr);
+		rtpe_redis_write->async_ctx = NULL;
+		if (redis_delete_async_context->errstr) {
+			rlog(LOG_ERROR, "redis_delete_async_context_connect error %d: %s",
+				redis_delete_async_context->err, redis_delete_async_context->errstr);
 		} else {
-			rlog(LOG_ERROR, "redis_async_context_disconnect error %d on context free: no errstr",
-				redis_notify_async_context->err);
+			rlog(LOG_ERROR, "redis_delete_async_context_connect error %d: no errstr",
+				redis_delete_async_context->err);
 		}
 	} else if (status == REDIS_OK) {
-		rlog(LOG_ERROR, "redis_async_context_disconnect initiated by user");
+		rlog(LOG_NOTICE, "redis_delete_async_context_connect initiated by user");
 	} else {
-		rlog(LOG_ERROR, "redis_async_context_disconnect invalid status code %d", status);
+		rlog(LOG_ERROR, "redis_delete_async_context_connect invalid status code %d", status);
 	}
 }
 
-int redis_async_context_alloc(void) {
-	struct redis *r = 0;
-
-	if (!rtpe_redis_notify) {
-		rlog(LOG_INFO, "redis_notify database is NULL.");
-		return -1;
+void redis_delete_async_context_disconnect(const redisAsyncContext *redis_delete_async_context, int status) {
+	rtpe_redis_write->async_ctx = NULL;
+	if (status == REDIS_ERR) {
+		if (redis_delete_async_context->errstr) {
+			rlog(LOG_ERROR, "redis_delete_async_context_disconnect error %d: %s",
+				redis_delete_async_context->err, redis_delete_async_context->errstr);
+		} else {
+			rlog(LOG_ERROR, "redis_delete_async_context_disconnect error %d: no errstr",
+				redis_delete_async_context->err);
+		}
+	} else if (status == REDIS_OK) {
+		rlog(LOG_NOTICE, "redis_delete_async_context_disconnect initiated by user");
+	} else {
+		rlog(LOG_ERROR, "redis_delete_async_context_disconnect invalid status code %d", status);
 	}
+}
 
-	// get redis_notify database
-	r = rtpe_redis_notify;
-	rlog(LOG_INFO, "Use Redis %s for notifications", endpoint_print_buf(&r->endpoint));
+void redis_notify_async_context_disconnect(const redisAsyncContext *redis_notify_async_context, int status) {
+	if (status == REDIS_ERR) {
+		if (redis_notify_async_context->errstr) {
+			rlog(LOG_ERROR, "redis_notify_async_context_disconnect error %d on context free: %s",
+				redis_notify_async_context->err, redis_notify_async_context->errstr);
+		} else {
+			rlog(LOG_ERROR, "redis_notify_async_context_disconnect error %d on context free: no errstr",
+				redis_notify_async_context->err);
+		}
+	} else if (status == REDIS_OK) {
+		rlog(LOG_NOTICE, "redis_notify_async_context_disconnect initiated by user");
+	} else {
+		rlog(LOG_ERROR, "redis_notify_async_context_disconnect invalid status code %d", status);
+	}
+}
+
+// connect_cb = connect callback, disconnect_cb = disconnect callback
+int redis_async_context_alloc(struct redis *r, void *connect_cb, void *disconnect_cb) {
+	// sanity checks
+	if (!r) {
+		rlog(LOG_ERROR, "redis_async_context_alloc: NULL r");
+		return -1;
+	} else {
+		rlog(LOG_DEBUG, "redis_async_context_alloc: Use Redis %s", endpoint_print_buf(&r->endpoint));
+	}
 
 	// alloc async context
-	rtpe_redis_notify_async_context = redisAsyncConnect(r->host, r->endpoint.port);
-	if (!rtpe_redis_notify_async_context) {
-		rlog(LOG_ERROR, "redis_notify_async_context can't create new");
-		return -1;
-	}
-	if (rtpe_redis_notify_async_context->err) {
-		rlog(LOG_ERROR, "redis_notify_async_context can't create new error: %s", rtpe_redis_notify_async_context->errstr);
+	r->async_ctx = redisAsyncConnect(r->host, r->endpoint.port);
+	if (!r->async_ctx) {
+		rlog(LOG_ERROR, "redis_async_context_alloc: can't create new");
 		return -1;
 	}
 
-	if (redisAsyncSetDisconnectCallback(rtpe_redis_notify_async_context, redis_async_context_disconnect) != REDIS_OK) {
-		rlog(LOG_ERROR, "redis_notify_async_context can't set disconnect callback");
+	if (r->async_ctx->err) {
+		rlog(LOG_ERROR, "redis_async_context_alloc: can't create new error: %s", r->async_ctx->errstr);
 		return -1;
 	}
 
+	// callbacks async context
+	if (redisAsyncSetConnectCallback(r->async_ctx, connect_cb) != REDIS_OK) {
+		rlog(LOG_ERROR, "redis_async_context_alloc: can't set connect callback");
+		return -1;
+	}
+
+	if (redisAsyncSetDisconnectCallback(r->async_ctx, disconnect_cb) != REDIS_OK) {
+		rlog(LOG_ERROR, "redis_async_context_alloc: can't set disconnect callback");
+		return -1;
+	}
+
+	rlog(LOG_DEBUG, "redis_async_context_alloc: Success");
+
 	return 0;
 }
 
-int redis_notify_event_base_action(enum event_base_action action) {
-	if (!rtpe_redis_notify_event_base && action!=EVENT_BASE_ALLOC) {
-		rlog(LOG_ERROR, "redis_notify_event_base is NULL on event base action %d", action);
+int redis_async_event_base_action(struct redis *r, enum event_base_action action) {
+	// sanity checks
+	if (!r) {
+		rlog(LOG_ERR, "redis_async_event_base_action: NULL r");
+		return -1;
+	} else {
+		rlog(LOG_DEBUG, "redis_async_event_base_action: Use Redis %s", endpoint_print_buf(&r->endpoint));
+	}
+
+	if (!r->async_ev && action != EVENT_BASE_ALLOC) {
+		rlog(LOG_NOTICE, "redis_async_event_base_action: async_ev is NULL on event base action %d", action);
 		return -1;
 	}
 
 	// exec event base action
 	switch (action) {
 		case EVENT_BASE_ALLOC:
-			rtpe_redis_notify_event_base = event_base_new();
-			if (!rtpe_redis_notify_event_base) {
-				rlog(LOG_ERROR, "Fail alloc redis_notify_event_base");
+			r->async_ev = event_base_new();
+			if (!r->async_ev) {
+				rlog(LOG_ERROR, "redis_async_event_base_action: Fail alloc async_ev");
 				return -1;
 			} else {
-				rlog(LOG_DEBUG, "Success alloc redis_notify_event_base");
+				rlog(LOG_DEBUG, "redis_async_event_base_action: Success alloc async_ev");
 			}
 			break;
 
 		case EVENT_BASE_FREE:
-			event_base_free(rtpe_redis_notify_event_base);
-			rlog(LOG_DEBUG, "Success free redis_notify_event_base");
+			event_base_free(r->async_ev);
+			rlog(LOG_DEBUG, "redis_async_event_base_action: Success free async_ev");
 			break;
 
 		case EVENT_BASE_LOOPBREAK:
-			if (event_base_loopbreak(rtpe_redis_notify_event_base)) {
-				rlog(LOG_ERROR, "Fail loopbreak redis_notify_event_base");
+			if (event_base_loopbreak(r->async_ev)) {
+				rlog(LOG_ERROR, "redis_async_event_base_action: Fail loopbreak async_ev");
 				return -1;
 			} else {
-				rlog(LOG_DEBUG, "Success loopbreak redis_notify_event_base");
+				rlog(LOG_DEBUG, "redis_async_event_base_action: Success loopbreak async_ev");
 			}
 			break;
 
 		default:
-			rlog(LOG_ERROR, "No event base action found: %d", action);
+			rlog(LOG_ERROR, "redis_async_event_base_action: No event base action found: %d", action);
 			return -1;
 	}
 
 	return 0;
 }
 
-int redis_notify_subscribe_action(enum subscribe_action action, int keyspace) {
-	if (!rtpe_redis_notify_async_context) {
+int redis_notify_subscribe_action(struct redis *r, enum subscribe_action action, int keyspace) {
+	if (!r->async_ctx) {
 		rlog(LOG_ERROR, "redis_notify_async_context is NULL on subscribe action");
 		return -1;
 	}
 
-	if (rtpe_redis_notify_async_context->err) {
-		rlog(LOG_ERROR, "redis_notify_async_context error on subscribe action: %s", rtpe_redis_notify_async_context->errstr);
+	if (r->async_ctx->err) {
+		rlog(LOG_ERROR, "redis_notify_async_context error on subscribe action: %s", r->async_ctx->errstr);
 		return -1;
 	}
 
 	switch (action) {
 	case SUBSCRIBE_KEYSPACE:
-		if (redisAsyncCommand(rtpe_redis_notify_async_context, on_redis_notification, NULL, "psubscribe __keyspace@%i__:*", keyspace) != REDIS_OK) {
+		if (redisAsyncCommand(r->async_ctx, on_redis_notification, NULL, "psubscribe __keyspace@%i__:*", keyspace) != REDIS_OK) {
 			rlog(LOG_ERROR, "Fail redisAsyncCommand on JSON SUBSCRIBE_KEYSPACE");
 			return -1;
 		}
 		break;
 	case UNSUBSCRIBE_KEYSPACE:
-		if (redisAsyncCommand(rtpe_redis_notify_async_context, on_redis_notification, NULL, "punsubscribe __keyspace@%i__:*", keyspace) != REDIS_OK) {
+		if (redisAsyncCommand(r->async_ctx, on_redis_notification, NULL, "punsubscribe __keyspace@%i__:*", keyspace) != REDIS_OK) {
 			rlog(LOG_ERROR, "Fail redisAsyncCommand on JSON UNSUBSCRIBE_KEYSPACE");
 			return -1;
 		}
 		break;
 	case UNSUBSCRIBE_ALL:
-		if (redisAsyncCommand(rtpe_redis_notify_async_context, on_redis_notification, NULL, "punsubscribe") != REDIS_OK) {
+		if (redisAsyncCommand(r->async_ctx, on_redis_notification, NULL, "punsubscribe") != REDIS_OK) {
 			rlog(LOG_ERROR, "Fail redisAsyncCommand on JSON UNSUBSCRIBE_ALL");
 			return -1;
 		}
@@ -523,33 +585,101 @@ int redis_notify_subscribe_action(enum subscribe_action action, int keyspace) {
 	return 0;
 }
 
-static int redis_notify(void) {
-	struct redis *r = 0;
+static int redis_delete_async(struct redis *r) {
+	// sanity checks
+	if (!r) {
+		rlog(LOG_ERROR, "redis_delete_async: Don't use Redis async deletions beacause no redis/redis_write.");
+		return -1 ;
+	}
+
+	// alloc new redis async context
+	if (r->async_ctx == NULL && redis_async_context_alloc(r, redis_delete_async_context_connect, redis_delete_async_context_disconnect) < 0) {
+		r->async_ctx = NULL;
+		rlog(LOG_ERROR, "redis_delete_async: Failed to alloc async_ctx");
+		return -1;
+	}
+
+	// attach event base
+	if (redisLibeventAttach(r->async_ctx, r->async_ev) == REDIS_ERR) {
+		if (r->async_ctx->err) {
+			rlog(LOG_ERROR, "redis_delete_async: redis_delete_async_context can't attach event base error: %s", r->async_ctx->errstr);
+		} else {
+			rlog(LOG_ERROR, "redis_delete_async: redis_delete_async_context can't attach event base");
+
+		}
+		return -1;
+	}
+
+	// commands
+	if (r->auth) {
+		if (redisAsyncCommand(r->async_ctx, NULL, NULL, "AUTH %s", r->auth) != REDIS_OK) {
+			rlog(LOG_ERROR, "redis_delete_async: Fail redisAsyncCommand on AUTH");
+			return -1;
+		}
+	} else {
+		if (redisAsyncCommand(r->async_ctx, NULL, NULL, "PING") != REDIS_OK) {
+			rlog(LOG_ERROR, "redis_delete_async: Fail redisAsyncCommand on PING");
+			return -1;
+		}
+	}
+
+	// delete commands
+	gchar *redis_command;
+	gint redis_command_total = 0;
+
+	mutex_lock(&r->async_lock);
+	while (!g_queue_is_empty(&r->async_queue)) {
+		redis_command_total++;
+		redis_command = g_queue_pop_head(&r->async_queue);
+
+		if (redisAsyncCommand(r->async_ctx, NULL, NULL, redis_command) != REDIS_OK) {
+			rlog(LOG_ERROR, "redis_delete_async: Fail redisAsyncCommand on DELETE");
+		}
+
+		g_free(redis_command);
+	}
+	mutex_unlock(&r->async_lock);
+
+	rlog(LOG_NOTICE, "redis_delete_async: Queued DELETE redisAsyncCommand total: %d", redis_command_total);
+
+	// dispatch event base => thread blocks here
+	if (event_base_dispatch(r->async_ev) < 0) {
+		rlog(LOG_ERROR, "redis_delete_async: Fail event_base_dispatch()");
+		return -1;
+	}
+
+	// loopbreak
+	redisAsyncDisconnect(r->async_ctx);
+	r->async_ctx = NULL;
+
+	return 0;
+}
+
+static int redis_notify(struct redis *r) {
 	GList *l;
 
-	if (!rtpe_redis_notify) {
+	if (!r) {
 		rlog(LOG_ERROR, "redis_notify database is NULL on redis_notify()");
 		return -1;
 	}
 
-	if (!rtpe_redis_notify_async_context) {
+	if (!r->async_ctx) {
 		rlog(LOG_ERROR, "redis_notify_async_context is NULL on redis_notify()");
 		return -1;
 	}
 
-	if (!rtpe_redis_notify_event_base) {
+	if (!r->async_ev) {
 		rlog(LOG_ERROR, "redis_notify_event_base is NULL on redis_notify()");
 		return -1;
 	}
 
 	// get redis_notify database
-	r = rtpe_redis_notify;
 	rlog(LOG_INFO, "Use Redis %s to subscribe to notifications", endpoint_print_buf(&r->endpoint));
 
 	// attach event base
-	if (redisLibeventAttach(rtpe_redis_notify_async_context, rtpe_redis_notify_event_base) == REDIS_ERR) {
-		if (rtpe_redis_notify_async_context->err) {
-			rlog(LOG_ERROR, "redis_notify_async_context can't attach event base error: %s", rtpe_redis_notify_async_context->errstr);
+	if (redisLibeventAttach(r->async_ctx, r->async_ev) == REDIS_ERR) {
+		if (r->async_ctx->err) {
+			rlog(LOG_ERROR, "redis_notify_async_context can't attach event base error: %s", r->async_ctx->errstr);
 		} else {
 			rlog(LOG_ERROR, "redis_notify_async_context can't attach event base");
 
@@ -557,8 +687,8 @@ static int redis_notify(void) {
 		return -1;
 	}
 
-	if (rtpe_redis_notify->auth) {
-		if (redisAsyncCommand(rtpe_redis_notify_async_context, on_redis_notification, NULL, "AUTH %s", rtpe_redis_notify->auth) != REDIS_OK) {
+	if (r->auth) {
+		if (redisAsyncCommand(r->async_ctx, on_redis_notification, NULL, "AUTH %s", r->auth) != REDIS_OK) {
 			rlog(LOG_ERROR, "Fail redisAsyncCommand on AUTH");
 			return -1;
 		}
@@ -567,12 +697,12 @@ static int redis_notify(void) {
 	// subscribe to the values in the configured keyspaces
 	rwlock_lock_r(&rtpe_config.config_lock);
 	for (l = rtpe_config.redis_subscribed_keyspaces.head; l; l = l->next) {
-		redis_notify_subscribe_action(SUBSCRIBE_KEYSPACE, GPOINTER_TO_UINT(l->data));
+		redis_notify_subscribe_action(r, SUBSCRIBE_KEYSPACE, GPOINTER_TO_INT(l->data));
 	}
 	rwlock_unlock_r(&rtpe_config.config_lock);
 
 	// dispatch event base => thread blocks here
-	if (event_base_dispatch(rtpe_redis_notify_event_base) < 0) {
+	if (event_base_dispatch(r->async_ev) < 0) {
 		rlog(LOG_ERROR, "Fail event_base_dispatch()");
 		return -1;
 	}
@@ -580,8 +710,35 @@ static int redis_notify(void) {
 	return 0;
 }
 
-static void redis_disconnect(void) {
-	rtpe_redis_notify_async_context = NULL;
+void redis_delete_async_loop(void *d) {
+	struct redis *r = NULL;
+
+	// sanity checks
+	r = rtpe_redis_write;
+	if (!r) {
+		rlog(LOG_ERROR, "redis_delete_async_loop: Don't use Redis async deletions beacause no redis/redis_write.");
+		return ;
+	}
+
+	r->async_last = rtpe_now.tv_sec;
+
+	// init libevent for pthread usage
+	if (evthread_use_pthreads() < 0) {
+		ilog(LOG_ERROR, "redis_delete_async_loop: evthread_use_pthreads failed.");
+		return ;
+	}
+
+	// alloc libevent base
+	if (redis_async_event_base_action(r, EVENT_BASE_ALLOC) < 0) {
+		rlog(LOG_ERROR, "redis_delete_async_loop: Failed to EVENT_BASE_ALLOC.");
+		return ;
+	}
+
+	// loop (almost) forever
+	while (!rtpe_shutdown) {
+		redis_delete_async(r);
+		sleep(1);
+	}
 }
 
 void redis_notify_loop(void *d) {
@@ -602,18 +759,18 @@ void redis_notify_loop(void *d) {
 	}
 
 	// alloc redis async context 
-	if (redis_async_context_alloc() < 0) {
+	if (redis_async_context_alloc(r, NULL, redis_notify_async_context_disconnect) < 0) {
 		return ;
 	}
 
 	// alloc event base
-	if (redis_notify_event_base_action(EVENT_BASE_ALLOC) < 0) {
+	if (redis_async_event_base_action(r, EVENT_BASE_ALLOC) < 0) {
 		return ;
 	}
 
 	// initial redis_notify
 	if (redis_check_conn(r) == REDIS_STATE_CONNECTED) {
-		redis_notify_return = redis_notify();
+		redis_notify_return = redis_notify(r);
 	}
 
 	// loop redis_notify => in case of lost connection
@@ -627,28 +784,23 @@ void redis_notify_loop(void *d) {
 		next_run = rtpe_now.tv_sec + seconds;
 
 		if (redis_check_conn(r) == REDIS_STATE_CONNECTED || redis_notify_return < 0) {
-			redis_disconnect();
+			r->async_ctx = NULL;
 			// alloc new redis async context upon redis breakdown
-			if (redis_async_context_alloc() < 0) {
+			if (redis_async_context_alloc(r, NULL, redis_notify_async_context_disconnect) < 0) {
 				continue;
 			}
 
 			// prepare notifications
-			redis_notify_return = redis_notify();
+			redis_notify_return = redis_notify(r);
 		}
 	}
 
-	// free libevent
-#if LIBEVENT_VERSION_NUMBER >= 0x02010100
-	libevent_global_shutdown();
-#endif
-
 	// unsubscribe notifications
-	redis_notify_subscribe_action(UNSUBSCRIBE_ALL, 0);
+	redis_notify_subscribe_action(r, UNSUBSCRIBE_ALL, 0);
 
 	// free async context
-	redisAsyncDisconnect(rtpe_redis_notify_async_context);
-	redis_disconnect();
+	redisAsyncDisconnect(r->async_ctx);
+	r->async_ctx = NULL;
 }
 
 struct redis *redis_new(const endpoint_t *ep, int db, const char *auth,
@@ -724,6 +876,7 @@ static void redis_count_err_and_disable(struct redis *r)
 
 /* must be called with r->lock held */
 static int redis_check_conn(struct redis *r) {
+	gettimeofday(&rtpe_now, NULL);
 
 	if ((r->state == REDIS_STATE_DISCONNECTED) && (r->restore_tick > rtpe_now.tv_sec)) {
 		ilog(LOG_WARNING, "Redis server %s is disabled. Don't try RE-Establishing for %ld more seconds",
@@ -737,6 +890,11 @@ static int redis_check_conn(struct redis *r) {
 	// try redis connection
 	if (r->ctx && redisCommandNR(r->ctx, "PING") == 0) {
 		// redis is connected
+		if (r->state == REDIS_STATE_DISCONNECTED) {
+			rlog(LOG_INFO, "RE-Established connection to Redis %s; PING works",
+				endpoint_print_buf(&r->endpoint));
+			r->state = REDIS_STATE_CONNECTED;
+		}
 		return REDIS_STATE_CONNECTED;
 	}
 
@@ -773,6 +931,16 @@ static void redis_delete_call_json(struct call *c, struct redis *r) {
 	redis_consume(r);
 }
 
+static void redis_delete_async_call_json(struct call *c, struct redis *r) {
+	gchar *redis_command;
+
+	redis_command = g_strdup_printf("SELECT %i", c->redis_hosted_db);
+	g_queue_push_tail(&r->async_queue, redis_command);
+
+	redis_command = g_strdup_printf("DEL " STR_FORMAT, STR_FMT(&c->callid));
+	g_queue_push_tail(&r->async_queue, redis_command);
+}
+
 INLINE void json_builder_add_string_value_uri_enc(JsonBuilder *builder, const char* tmp, int len) {
 	char enc[len * 3 + 1];
 	str_uri_encode_len(enc, tmp, len);
@@ -884,7 +1052,7 @@ static int redis_hash_get_str(str *out, const struct redis_hash *h, const char *
 
 /* we can do this because this happens during startup in a single thread */
 static atomic64 strtoa64(const char *c, char **endp, int base) {
-	u_int64_t u;
+	uint64_t u;
 	atomic64 ret;
 
 	u = strtoull(c, endp, base);
@@ -902,15 +1070,15 @@ define_get_int_type(time_t, time_t, strtoull);
 define_get_int_type(timeval, struct timeval, strtotimeval);
 define_get_int_type(int, int, strtol);
 define_get_int_type(unsigned, unsigned int, strtol);
-//define_get_int_type(u16, u_int16_t, strtol);
-//define_get_int_type(u64, u_int64_t, strtoull);
+//define_get_int_type(u16, uint16_t, strtol);
+//define_get_int_type(u64, uint64_t, strtoull);
 define_get_int_type(a64, atomic64, strtoa64);
 
 define_get_type_format(str, str);
 define_get_type_format(int, int);
 //define_get_type_format(unsigned, unsigned int);
-//define_get_type_format(u16, u_int16_t);
-//define_get_type_format(u64, u_int64_t);
+//define_get_type_format(u16, uint16_t);
+//define_get_type_format(u64, uint64_t);
 define_get_type_format(a64, atomic64);
 
 static int redis_hash_get_c_buf_fn(unsigned char *out, size_t len, const struct redis_hash *h,
@@ -1180,6 +1348,9 @@ static int redis_sfds(struct call *c, struct redis_list *sfds) {
 		set_tos(sock, c->tos);
 		sfd = stream_fd_new(sock, c, loc);
 
+		if (redis_hash_get_sdes_params1(&sfd->crypto.params, rh, "") == -1)
+			return -1;
+
 		sfds->ptrs[i] = sfd;
 	}
 	return 0;
@@ -1212,6 +1383,8 @@ static int redis_streams(struct call *c, struct redis_list *streams) {
 			return -1;
 		if (redis_hash_get_stats(&ps->stats, rh, "stats"))
 			return -1;
+		if (redis_hash_get_sdes_params1(&ps->crypto.params, rh, "") == -1)
+			return -1;
 
 		streams->ptrs[i] = ps;
 
@@ -1220,7 +1393,7 @@ static int redis_streams(struct call *c, struct redis_list *streams) {
 	return 0;
 }
 
-static int redis_tags(struct call *c, struct redis_list *tags) {
+static int redis_tags(struct call *c, struct redis_list *tags, JsonReader *root_reader) {
 	unsigned int i;
 	int ii;
 	struct redis_hash *rh;
@@ -1248,6 +1421,9 @@ static int redis_tags(struct call *c, struct redis_list *tags) {
 		if (!redis_hash_get_int(&ii, rh, "block_media"))
 			ml->block_media = ii ? 1 : 0;
 
+		if (json_build_ssrc(ml, root_reader))
+			return -1;
+
 		tags->ptrs[i] = ml;
 	}
 
@@ -1271,7 +1447,7 @@ static struct rtp_payload_type *rbl_cb_plts_g(str *s, GQueue *q, struct redis_li
 }
 static int rbl_cb_plts_r(str *s, GQueue *q, struct redis_list *list, void *ptr) {
 	struct call_media *med = ptr;
-	__rtp_payload_type_add_recv(med, rbl_cb_plts_g(s, q, list, ptr));
+	__rtp_payload_type_add_recv(med, rbl_cb_plts_g(s, q, list, ptr), 0);
 	return 0;
 }
 static int rbl_cb_plts_s(str *s, GQueue *q, struct redis_list *list, void *ptr) {
@@ -1533,16 +1709,21 @@ static int json_link_maps(struct call *c, struct redis_list *maps,
 	return 0;
 }
 
-static int json_build_ssrc(struct call *c, JsonReader *root_reader) {
-	if (!json_reader_read_member(root_reader, "ssrc_table"))
-		return -1;
+static int json_build_ssrc(struct call_monologue *ml, JsonReader *root_reader) {
+	char tmp[2048];
+	snprintf(tmp, sizeof(tmp), "ssrc_table-%u", ml->unique_id);
+	if (!json_reader_read_member(root_reader, "ssrc_table")) {
+		// non-fatal for backwards compatibility
+		json_reader_end_member(root_reader);
+		return 0;
+	}
 	int nmemb = json_reader_count_elements(root_reader);
 	for (int jidx=0; jidx < nmemb; ++jidx) {
 		if (!json_reader_read_element(root_reader, jidx))
 			return -1;
 
-		u_int32_t ssrc = json_reader_get_ll(root_reader, "ssrc");
-		struct ssrc_entry_call *se = get_ssrc(ssrc, c->ssrc_hash);
+		uint32_t ssrc = json_reader_get_ll(root_reader, "ssrc");
+		struct ssrc_entry_call *se = get_ssrc(ssrc, ml->ssrc_hash);
 		se->input_ctx.srtp_index = json_reader_get_ll(root_reader, "in_srtp_index");
 		se->input_ctx.srtcp_index = json_reader_get_ll(root_reader, "in_srtcp_index");
 		payload_tracker_add(&se->input_ctx.tracker, json_reader_get_ll(root_reader, "in_payload_type"));
@@ -1563,6 +1744,7 @@ static void json_restore_call(struct redis *r, const str *callid, int foreign) {
 	struct redis_list tags, sfds, streams, medias, maps;
 	struct call *c = NULL;
 	str s, id, meta;
+	time_t last_signal;
 
 	const char *err = 0;
 	int i;
@@ -1588,13 +1770,25 @@ static void json_restore_call(struct redis *r, const str *callid, int foreign) {
 	if (!c)
 		goto err1;
 
-	err = "call already exists";
-	if (c->last_signal)
-		goto err2;
 	err = "'call' data incomplete";
-
 	if (json_get_hash(&call, "json", -1, root_reader))
 		goto err2;
+
+	err = "missing 'last signal' timestamp";
+	if (redis_hash_get_time_t(&last_signal, &call, "last_signal"))
+		goto err3;
+
+	if (c->last_signal) {
+		err = NULL;
+		// is the call we're loading newer than the one we have?
+		if (last_signal > c->last_signal) {
+			// switch ownership
+			call_make_own_foreign(c, foreign);
+			c->last_signal = last_signal;
+		}
+		goto err3; // no error, just bail
+	}
+
 	err = "'tags' incomplete";
 	if (json_get_list_hash(&tags, "tag", &call, "num_tags", root_reader))
 		goto err3;
@@ -1614,9 +1808,7 @@ static void json_restore_call(struct redis *r, const str *callid, int foreign) {
 	err = "missing 'created' timestamp";
 	if (redis_hash_get_timeval(&c->created, &call, "created"))
 		goto err8;
-	err = "missing 'last signal' timestamp";
-	if (redis_hash_get_time_t(&c->last_signal, &call, "last_signal"))
-		goto err8;
+	c->last_signal = last_signal;
 	if (redis_hash_get_int(&i, &call, "tos"))
 		c->tos = 184;
 	else
@@ -1643,7 +1835,7 @@ static void json_restore_call(struct redis *r, const str *callid, int foreign) {
 	if (redis_streams(c, &streams))
 		goto err8;
 	err = "failed to create tags";
-	if (redis_tags(c, &tags))
+	if (redis_tags(c, &tags, root_reader))
 		goto err8;
 	err = "failed to create medias";
 	if (json_medias(c, &medias, root_reader))
@@ -1667,9 +1859,6 @@ static void json_restore_call(struct redis *r, const str *callid, int foreign) {
 	err = "failed to link maps";
 	if (json_link_maps(c, &maps, &sfds, root_reader))
 		goto err8;
-	err = "failed to restore SSRC table";
-	if (json_build_ssrc(c, root_reader))
-		goto err8;
 
 	// presence of this key determines whether we were recording at all
 	if (!redis_hash_get_str(&s, &call, "recording_meta_prefix")) {
@@ -1707,10 +1896,15 @@ err1:
 				err);
 		if (c) 
 			call_destroy(c);
-		else {
-			mutex_lock(&rtpe_redis_write->lock);
-			redisCommandNR(rtpe_redis_write->ctx, "DEL " PB, STR(callid));
-			mutex_unlock(&rtpe_redis_write->lock);
+
+		mutex_lock(&rtpe_redis_write->lock);
+		redisCommandNR(rtpe_redis_write->ctx, "DEL " PB, STR(callid));
+		mutex_unlock(&rtpe_redis_write->lock);
+
+		if (rtpe_redis_notify) {
+			mutex_lock(&rtpe_redis_notify->lock);
+			redisCommandNR(rtpe_redis_notify->ctx, "DEL " PB, STR(callid));
+			mutex_unlock(&rtpe_redis_notify->lock);
 		}
 	}
 	if (c)
@@ -1720,6 +1914,7 @@ err1:
 struct thread_ctx {
 	GQueue r_q;
 	mutex_t r_m;
+	int foreign;
 };
 
 static void restore_thread(void *call_p, void *ctx_p) {
@@ -1735,14 +1930,14 @@ static void restore_thread(void *call_p, void *ctx_p) {
 	r = g_queue_pop_head(&ctx->r_q);
 	mutex_unlock(&ctx->r_m);
 
-	json_restore_call(r, &callid, 0);
+	json_restore_call(r, &callid, ctx->foreign);
 
 	mutex_lock(&ctx->r_m);
 	g_queue_push_tail(&ctx->r_q, r);
 	mutex_unlock(&ctx->r_m);
 }
 
-int redis_restore(struct redis *r) {
+int redis_restore(struct redis *r, int foreign, int db) {
 	redisReply *calls = NULL, *call;
 	int i, ret = -1;
 	GThreadPool *gtp;
@@ -1751,7 +1946,8 @@ int redis_restore(struct redis *r) {
 	if (!r)
 		return 0;
 
-	rtpe_config.common.log_level |= LOG_FLAG_RESTORE;
+	for (unsigned int i = 0; i < num_log_levels; i++)
+		rtpe_config.common.log_levels[i] |= LOG_FLAG_RESTORE;
 
 	rlog(LOG_DEBUG, "Restoring calls from Redis...");
 
@@ -1762,10 +1958,18 @@ int redis_restore(struct redis *r) {
 		ret = 0;
 		goto err;
 	}
-	mutex_unlock(&r->lock);
+	if (db != -1)
+		redis_select_db(r, db);
 
 	calls = redis_get(r, REDIS_REPLY_ARRAY, "KEYS *");
 
+	if (db != -1)
+		redis_select_db(r, r->db);
+	else
+		db = r->db;
+
+	mutex_unlock(&r->lock);
+
 	if (!calls) {
 		rlog(LOG_ERR, "Could not retrieve call list from Redis: %s",
 				r->ctx ? r->ctx->errstr : "No redis context");
@@ -1774,9 +1978,10 @@ int redis_restore(struct redis *r) {
 
 	mutex_init(&ctx.r_m);
 	g_queue_init(&ctx.r_q);
+	ctx.foreign = foreign;
 	for (i = 0; i < rtpe_config.redis_num_threads; i++)
 		g_queue_push_tail(&ctx.r_q,
-				redis_new(&r->endpoint, r->db, r->auth, r->role, r->no_redis_required));
+				redis_new(&r->endpoint, db, r->auth, r->role, r->no_redis_required));
 	gtp = g_thread_pool_new(restore_thread, &ctx, rtpe_config.redis_num_threads, TRUE, NULL);
 
 	for (i = 0; i < calls->elements; i++) {
@@ -1798,7 +2003,9 @@ int redis_restore(struct redis *r) {
 	freeReplyObject(calls);
 
 err:
-	rtpe_config.common.log_level &= ~LOG_FLAG_RESTORE;
+	for (unsigned int i = 0; i < num_log_levels; i++)
+		if (rtpe_config.common.log_levels[i] > 0)
+			rtpe_config.common.log_levels[i] &= ~LOG_FLAG_RESTORE;
 	return ret;
 }
 
@@ -1828,6 +2035,24 @@ err:
 #define JSON_SET_SIMPLE_CSTR(a,d) JSON_SET_SIMPLE_LEN(a, strlen(d), d)
 #define JSON_SET_SIMPLE_STR(a,d) JSON_SET_SIMPLE_LEN(a, (d)->len, (d)->s)
 
+static void json_update_crypto_params(JsonBuilder *builder, const char *key, struct crypto_params *p) {
+	char tmp[2048];
+
+	if (!p->crypto_suite)
+		return;
+
+	JSON_SET_NSTRING_CSTR("%s-crypto_suite", key, p->crypto_suite->name);
+	JSON_SET_NSTRING_LEN("%s-master_key", key, sizeof(p->master_key), (char *) p->master_key);
+	JSON_SET_NSTRING_LEN("%s-master_salt", key, sizeof(p->master_salt), (char *) p->master_salt);
+
+	JSON_SET_NSTRING("%s-unenc-srtp", key, "%i", p->session_params.unencrypted_srtp);
+	JSON_SET_NSTRING("%s-unenc-srtcp", key, "%i", p->session_params.unencrypted_srtcp);
+	JSON_SET_NSTRING("%s-unauth-srtp", key, "%i", p->session_params.unauthenticated_srtp);
+
+	if (p->mki)
+		JSON_SET_NSTRING_LEN("%s-mki", key, p->mki_len, (char *) p->mki);
+}
+
 static int json_update_sdes_params(JsonBuilder *builder, const char *pref,
 		unsigned int unique_id,
 		const char *k, GQueue *q)
@@ -1845,17 +2070,7 @@ static int json_update_sdes_params(JsonBuilder *builder, const char *pref,
 			return -1;
 
 		JSON_SET_NSTRING("%s_tag", key, "%u", cps->tag);
-		JSON_SET_NSTRING_CSTR("%s-crypto_suite",key,p->crypto_suite->name);
-		JSON_SET_NSTRING_LEN("%s-master_key",key, sizeof(p->master_key), (char *) p->master_key);
-		JSON_SET_NSTRING_LEN("%s-master_salt",key, sizeof(p->master_salt), (char *) p->master_salt);
-
-		JSON_SET_NSTRING("%s-unenc-srtp",key,"%i",p->session_params.unencrypted_srtp);
-		JSON_SET_NSTRING("%s-unenc-srtcp",key,"%i",p->session_params.unencrypted_srtcp);
-		JSON_SET_NSTRING("%s-unauth-srtp",key,"%i",p->session_params.unauthenticated_srtp);
-
-		if (p->mki) {
-			JSON_SET_NSTRING_LEN("%s-mki",key, p->mki_len, (char *) p->mki);
-		}
+		json_update_crypto_params(builder, key, p);
 
 		snprintf(keybuf, sizeof(keybuf), "%s-%u", k, iter++);
 		key = keybuf;
@@ -1940,6 +2155,8 @@ char* redis_encode_json(struct call *c) {
 				JSON_SET_SIMPLE("local_intf_uid","%u",sfd->local_intf->unique_id);
 				JSON_SET_SIMPLE("stream","%u",sfd->stream->unique_id);
 
+				json_update_crypto_params(builder, "", &sfd->crypto.params);
+
 			}
 			json_builder_end_object (builder);
 
@@ -1971,6 +2188,7 @@ char* redis_encode_json(struct call *c) {
 				JSON_SET_SIMPLE("stats-bytes","%" PRIu64, atomic64_get(&ps->stats.bytes));
 				JSON_SET_SIMPLE("stats-errors","%" PRIu64, atomic64_get(&ps->stats.errors));
 
+				json_update_crypto_params(builder, "", &ps->crypto.params);
 			}
 
 			json_builder_end_object (builder);
@@ -2065,6 +2283,33 @@ char* redis_encode_json(struct call *c) {
 				JSON_ADD_STRING("%u",media->unique_id);
 			}
 			json_builder_end_array (builder);
+
+			// SSRC table dump
+			rwlock_lock_r(&ml->ssrc_hash->lock);
+			k = g_hash_table_get_values(ml->ssrc_hash->ht);
+			snprintf(tmp, sizeof(tmp), "ssrc_table-%u", ml->unique_id);
+			json_builder_set_member_name(builder, tmp);
+			json_builder_begin_array (builder);
+			for (m = k; m; m = m->next) {
+				struct ssrc_entry_call *se = m->data;
+				json_builder_begin_object (builder);
+
+				JSON_SET_SIMPLE("ssrc","%" PRIu32, se->h.ssrc);
+				// XXX use function for in/out
+				JSON_SET_SIMPLE("in_srtp_index","%" PRIu64, se->input_ctx.srtp_index);
+				JSON_SET_SIMPLE("in_srtcp_index","%" PRIu64, se->input_ctx.srtcp_index);
+				JSON_SET_SIMPLE("in_payload_type","%i", se->input_ctx.tracker.most[0]);
+				JSON_SET_SIMPLE("out_srtp_index","%" PRIu64, se->output_ctx.srtp_index);
+				JSON_SET_SIMPLE("out_srtcp_index","%" PRIu64, se->output_ctx.srtcp_index);
+				JSON_SET_SIMPLE("out_payload_type","%i", se->output_ctx.tracker.most[0]);
+				// XXX add rest of info
+
+				json_builder_end_object (builder);
+			}
+			json_builder_end_array (builder);
+
+			g_list_free(k);
+			rwlock_unlock_r(&ml->ssrc_hash->lock);
 		}
 
 
@@ -2184,31 +2429,6 @@ char* redis_encode_json(struct call *c) {
 			json_builder_end_array (builder);
 		}
 
-		// SSRC table dump
-		rwlock_lock_r(&c->ssrc_hash->lock);
-		k = g_hash_table_get_values(c->ssrc_hash->ht);
-		json_builder_set_member_name(builder, "ssrc_table");
-		json_builder_begin_array (builder);
-		for (m = k; m; m = m->next) {
-			struct ssrc_entry_call *se = m->data;
-			json_builder_begin_object (builder);
-
-			JSON_SET_SIMPLE("ssrc","%" PRIu32, se->h.ssrc);
-			// XXX use function for in/out
-			JSON_SET_SIMPLE("in_srtp_index","%" PRIu64, se->input_ctx.srtp_index);
-			JSON_SET_SIMPLE("in_srtcp_index","%" PRIu64, se->input_ctx.srtcp_index);
-			JSON_SET_SIMPLE("in_payload_type","%i", se->input_ctx.tracker.most[0]);
-			JSON_SET_SIMPLE("out_srtp_index","%" PRIu64, se->output_ctx.srtp_index);
-			JSON_SET_SIMPLE("out_srtcp_index","%" PRIu64, se->output_ctx.srtcp_index);
-			JSON_SET_SIMPLE("out_payload_type","%i", se->output_ctx.tracker.most[0]);
-			// XXX add rest of info
-
-			json_builder_end_object (builder);
-		}
-		json_builder_end_array (builder);
-
-		g_list_free(k);
-		rwlock_unlock_r(&c->ssrc_hash->lock);
 	}
 	json_builder_end_object (builder);
 
@@ -2231,6 +2451,8 @@ void redis_update_onekey(struct call *c, struct redis *r) {
 
 	if (!r)
 		return;
+	if (c->foreign_call)
+		return;
 
 	mutex_lock(&r->lock);
 	// coverity[sleep : FALSE]
@@ -2244,7 +2466,7 @@ void redis_update_onekey(struct call *c, struct redis *r) {
 	redis_expires_s = rtpe_config.redis_expires_secs;
 
 	c->redis_hosted_db = r->db;
-	if (redisCommandNR(r->ctx, "SELECT %i", c->redis_hosted_db)) {
+	if (redis_select_db(r, c->redis_hosted_db)) {
 		rlog(LOG_ERR, " >>>>>>>>>>>>>>>>> Redis error.");
 		goto err;
 	}
@@ -2276,9 +2498,21 @@ err:
 
 /* must be called lock-free */
 void redis_delete(struct call *c, struct redis *r) {
+	int delete_async = rtpe_config.redis_delete_async;
+	rlog(LOG_DEBUG, "Redis delete_async=%d", delete_async);
+
 	if (!r)
 		return;
 
+	if (delete_async) {
+		mutex_lock(&r->async_lock);
+		rwlock_lock_r(&c->master_lock);
+		redis_delete_async_call_json(c, r);
+		rwlock_unlock_r(&c->master_lock);
+		mutex_unlock(&r->async_lock);
+		return;
+	}
+
 	mutex_lock(&r->lock);
 	// coverity[sleep : FALSE]
 	if (redis_check_conn(r) == REDIS_STATE_DISCONNECTED) {
@@ -2287,7 +2521,7 @@ void redis_delete(struct call *c, struct redis *r) {
 	}
 	rwlock_lock_r(&c->master_lock);
 
-	if (redisCommandNR(r->ctx, "SELECT %i", c->redis_hosted_db))
+	if (redis_select_db(r, c->redis_hosted_db))
 		goto err;
 
 	redis_delete_call_json(c, r);
