@@ -8,6 +8,7 @@
 #include <assert.h>
 #include <spandsp/t30.h>
 #include <spandsp/logging.h>
+#include "spandsp_logging.h"
 #include "codec.h"
 #include "call.h"
 #include "log.h"
@@ -76,14 +77,14 @@ static void g_string_null_extend(GString *s, size_t len) {
 	memset(s->str + oldb, 0, newb);
 }
 
-static void spandsp_logging_func(int level, const char *text) {
+static void spandsp_logging_func(SPAN_LOG_ARGS) {
 	if (level <= SPAN_LOG_PROTOCOL_ERROR)
 		level = LOG_ERR;
 	else if (level <= SPAN_LOG_PROTOCOL_WARNING)
 		level = LOG_WARN;
 	else
 		level = LOG_DEBUG;
-	ilog(level, "SpanDSP: %s", text);
+	ilogs(spandsp, level, "SpanDSP: %s", text);
 }
 
 
@@ -412,8 +413,8 @@ int t38_gateway_pair(struct call_media *t38_media, struct call_media *pcm_media,
 	t38_set_max_datagram_size(t38, opts.max_ifp);
 
 	logging_state_t *ls = t38_gateway_get_logging_state(tg->gw);
-	span_log_set_message_handler(ls, spandsp_logging_func);
-	span_log_set_level(ls, span_log_level_map(get_log_level()));
+	my_span_set_log(ls, spandsp_logging_func);
+	span_log_set_level(ls, span_log_level_map(get_log_level(spandsp)));
 
 	packet_sequencer_init(&tg->sequencer, (GDestroyNotify) __udptl_packet_free);
 	tg->sequencer.seq = 0;
@@ -567,11 +568,11 @@ int t38_gateway_input_udptl(struct t38_gateway *tg, const str *buf) {
 		return 0;
 
 	if (buf->len < 4) {
-		ilog(LOG_INFO | LOG_FLAG_LIMIT, "Ignoring short UDPTL packet (%i bytes)", buf->len);
+		ilog(LOG_INFO | LOG_FLAG_LIMIT, "Ignoring short UDPTL packet (%zu bytes)", buf->len);
 		return 0;
 	}
 
-	ilog(LOG_DEBUG, "Processing %i UDPTL bytes", buf->len);
+	ilog(LOG_DEBUG, "Processing %zu UDPTL bytes", buf->len);
 
 	str s = *buf;
 	str piece;
@@ -579,20 +580,20 @@ int t38_gateway_input_udptl(struct t38_gateway *tg, const str *buf) {
 	// get seq num
 	uint16_t seq;
 	if (str_shift_ret(&s, 2, &piece))
-		goto err;
+		goto err_nolock;
 	seq = ntohs(*((uint16_t *) piece.s));
 
 	err = "Invalid primary UDPTL packet";
 	if (__get_udptl(&piece, &s))
 		goto err;
 
-	ilog(LOG_DEBUG, "Received primary IFP packet, len %i, seq %i", piece.len, seq);
+	ilog(LOG_DEBUG, "Received primary IFP packet, len %zu, seq %i", piece.len, seq);
 	str primary = piece;
 	up = __make_udptl_packet(&primary, seq);
 
 	err = "Error correction mode byte missing";
 	if (str_shift_ret(&s, 1, &piece))
-		goto err;
+		goto err_nolock;
 	char fec = piece.s[0];
 
 	mutex_lock(&tg->lock);
@@ -635,7 +636,7 @@ int t38_gateway_input_udptl(struct t38_gateway *tg, const str *buf) {
 			// ignore zero-length packets
 			if (!piece.len)
 				continue;
-			ilog(LOG_DEBUG, "Received secondary IFP packet, len %i, seq %i", piece.len,
+			ilog(LOG_DEBUG, "Received secondary IFP packet, len %zu, seq %i", piece.len,
 					seq - 1 - i);
 			up = __make_udptl_packet(&piece, seq - 1 - i);
 			packet_sequencer_insert(&tg->sequencer, &up->p);
@@ -743,7 +744,7 @@ seq_ok:;
 		if (!up)
 			break;
 
-		ilog(LOG_DEBUG, "Processing %i IFP bytes, seq %i", up->s->len, up->p.seq);
+		ilog(LOG_DEBUG, "Processing %zu IFP bytes, seq %i", up->s->len, up->p.seq);
 
 		t38_core_rx_ifp_packet(t38, (uint8_t *) up->s->s, up->s->len, up->p.seq);
 
@@ -755,6 +756,8 @@ out:
 	return 0;
 
 err:
+	mutex_unlock(&tg->lock);
+err_nolock:
 	if (err)
 		ilog(LOG_ERR | LOG_FLAG_LIMIT, "Failed to process UDPTL/T.38/IFP packet: %s", err);
 	if (up)
@@ -774,7 +777,7 @@ void t38_gateway_stop(struct t38_gateway *tg) {
 
 
 void t38_init(void) {
-	span_set_message_handler(NULL);
+	my_span_mh(NULL);
 }
 
 
